from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, Union, cast

from psygnal import Signal
from typing_extensions import get_args, get_origin

from magicgui.types import Undefined, _Undefined

from ._widget import Widget

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.widgets import protocols

    from ._widget import WidgetKwargs

T = TypeVar("T")


class ValueWidget(Widget, Generic[T]):
    """Widget with a value, Wraps ValueWidgetProtocol.

    Parameters
    ----------
    value : Any, optional
        The starting value for the widget.
    bind : Callable[[ValueWidget], Any] | Any, optional
        A value or callback to bind this widget. If provided, whenever
        [`widget.value`][magicgui.widgets.bases.ValueWidget.value] is
        accessed, the value provided here will be returned instead. `bind` may be a
        callable, in which case `bind(self)` will be returned (i.e. your bound callback
        must accept a single parameter, which is this widget instance).
    nullable : bool, optional
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.ValueWidgetProtocol
    changed = Signal(object, description="Emitted when the widget value changes.")
    null_value: Any = None

    def __init__(
        self,
        value: T | _Undefined = Undefined,
        *,
        bind: T | Callable[[ValueWidget], T] | _Undefined = Undefined,
        nullable: bool = False,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        self._nullable = nullable
        self._bound_value = bind
        self._call_bound: bool = True
        super().__init__(**base_widget_kwargs)
        if value is not Undefined:
            self.value = cast(T, value)
        if self._bound_value is not Undefined and "visible" not in base_widget_kwargs:
            self.hide()

    def _post_init(self) -> None:
        super()._post_init()
        self._widget._mgui_bind_change_callback(self._on_value_change)

    def _on_value_change(self, value: T | None = None) -> None:
        """Called when the widget value changes."""
        if value is self.null_value and not self._nullable:
            return
        self.changed.emit(value)

    def get_value(self) -> T:
        """Callable version of `self.value`.

        The main API is to use `self.value`, however, this is here in order to provide
        an escape hatch if trying to access the widget's value inside of a callback
        bound to self._bound_value.
        """
        return cast(T, self._widget._mgui_get_value())

    @property
    def value(self) -> T:
        """Return current value of the widget.  This may be interpreted by backends."""
        if self._bound_value is not Undefined:
            if callable(self._bound_value) and self._call_bound:
                try:
                    return self._bound_value(self)
                except RecursionError as e:
                    raise RuntimeError(
                        "RecursionError in callback bound to "
                        f"<{self.widget_type!r} name={self.name!r}>. If you need to "
                        "access `widget.value` in your bound callback, use "
                        "`widget.get_value()`"
                    ) from e
            return cast(T, self._bound_value)
        return self.get_value()

    @value.setter
    def value(self, value: T) -> None:
        return self._widget._mgui_set_value(value)

    def __repr__(self) -> str:
        """Return representation of widget of instance."""
        try:
            val = self.value if self._bound_value is Undefined else self._bound_value
            return (
                f"{self.widget_type}(value={val!r}, "
                f"annotation={self.annotation!r}, name={self.name!r})"
            )
        except AttributeError:  # pragma: no cover
            return f"<Uninitialized {self.widget_type}>"

    def bind(self, value: T | Callable[[ValueWidget], T], call: bool = True) -> None:
        """Binds ``value`` to self.value.

        If a value is bound to this widget, then whenever `widget.value` is accessed,
        the value provided here will be returned.  ``value`` can be a callable, in which
        case ``value(self)`` will be returned (i.e. your callback must accept a single
        parameter, which is this widget instance.).

        If you provide a callable and you *don't* want it to be called (but rather just
        returned as a callable object, then use ``call=False`` when binding your value.

        Note: if you need to access the "original" ``widget.value`` within your
        callback, please use ``widget.get_value()`` instead of the ``widget.value``
        property, in order to avoid a RuntimeError.

        Parameters
        ----------
        value : Any
            The value (or callback) to return when accessing this widget's value.
        call : bool, optional
            If ``value`` is a callable and ``call`` is ``True``, the callback will be
            called as ``callback(self)`` when accessing ``self.value``.  If ``False``,
            the callback will simply be returned as a callable object, by default,
            ``True``.
        """
        self._call_bound = call
        self._bound_value = value

    def unbind(self) -> None:
        """Unbinds any bound values. (see ``ValueWidget.bind``)."""
        self._bound_value = Undefined

    @property
    def annotation(self) -> Any:
        """Return type annotation for the parameter represented by the widget.

        ForwardRefs will be resolve when setting the annotation.
        If the widget is nullable (had a type annototation of Optional[Type]),
        annotation will return the first argument in the Optional clause.
        """
        annotation = Widget.annotation.fget(self)  # type: ignore
        if self._nullable and get_origin(annotation) is Union:
            return get_args(annotation)[0]
        return annotation

    @annotation.setter
    def annotation(self, value: Any) -> None:
        Widget.annotation.fset(self, value)  # type: ignore



"""# Table widget

Demonstrating a few ways to input tables.
"""

import numpy as np

from magicgui.widgets import Table

# all of these are valid data types
dict_of_lists = {"col_1": [1, 4], "col_2": [2, 5], "col_3": [3, 6]}
# column-dict-of-row-dicts
dict_of_dict = {
    "col_1": {"r1": 1, "r2": 4},
    "col_2": {"r1": 2, "r2": 5},
    "col_3": {"r1": 3, "r2": 6},
}
# list-of-lists
list_of_list = [[1, 2, 3], [4, 5, 6]]
# Records: List-of-column-dict
list_of_records = [
    {"col_1": 1, "col_2": 2, "col_3": 3},
    {"col_1": 4, "col_2": 5, "col_3": 6},
]

# 3-tuple of data, index, column
data_index_column_tuple = (([[1, 2, 3], [4, 5, 6]], ("r1", "r2"), ("c1", "c2", "c3")),)
# split-dict
split_dict = {
    "data": [[1, 2, 3], [4, 5, 6]],
    "index": ("r1", "r2"),
    "columns": ("c1", "c2", "c3"),
}

table = Table(value=dict_of_lists)

# it behaves like a dict:
table["new_col"] = [5, 5]
assert table.pop("new_col") == [5, 5]
# keys and items have both regular (column) and "row" modes
col_item_view = table.items()  # iterate col_header/column
row_item_view = table.items("row")  # iterate row_header/row

# we can just call dict() to get back our dict of lists
assert dict(table) == dict_of_lists
# or use one of many other exports in `to_dict`
assert table.to_dict("records") == list_of_records

# change headers
table.row_headers = ("row1", "row2")
table.column_headers = ("a", "b", "c")

# setting value clears and resets the table:
table.value = np.arange(18).reshape(6, 3)
# we can get/set/delete the 2D data table using numpy-style indexing:
# get every other row
assert table.data[::2] == [[0, 1, 2], [6, 7, 8], [12, 13, 14]]
# set every other column in the 3rd row
table.data[2, ::2] = [99, 99]

# export to numpy or pandas
# table.data.to_numpy()
# table.to_dataframe()

# the table.changed event emits a dict of information on any cell change
# e.g. {'data': 'sdfg', 'row': 1, 'column': 0, 'column_header': '1', 'row_header': '1'}
table.changed.connect(print)
table.show(run=True)



"""The FunctionGui class is a Container subclass designed to represent a function.

The core `magicgui` decorator returns an instance of a FunctionGui widget.
"""

from __future__ import annotations

import inspect
import re
from collections import deque
from contextlib import contextmanager
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    NoReturn,
    TypeVar,
    cast,
)

from psygnal import Signal

from magicgui._type_resolution import resolve_single_type
from magicgui.signature import MagicSignature, magic_signature
from magicgui.widgets import Container, MainWindow, ProgressBar, PushButton
from magicgui.widgets.bases import ValueWidget

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import ParamSpec

    from magicgui.application import Application, AppRef  # noqa: F401
    from magicgui.widgets import TextEdit
    from magicgui.widgets.protocols import ContainerProtocol, MainWindowProtocol

    _P = ParamSpec("_P")
else:
    _P = TypeVar("_P")  # easier runtime dependency than ParamSpec


def _inject_tooltips_from_docstrings(
    docstring: str | None, sig: MagicSignature
) -> None:
    """Update `sig` gui options with tooltips extracted from `docstring`."""
    from docstring_parser import parse

    if not docstring:
        return

    doc_params: dict[str, str | None] = {
        p.arg_name: p.description for p in parse(docstring).params
    }
    if not doc_params:
        return

    # deal with the (numpydocs) case when there are multiple parameters separated
    # by a comma
    for k, v in list(doc_params.items()):
        if "," in k:
            for split_key in k.split(","):
                doc_params[split_key.strip()] = v
            del doc_params[k]
    for name, description in doc_params.items():
        # this is to catch potentially bad arg_name parsing in docstring_parser
        # if using napoleon style google docstringss
        argname = name.split(" ", maxsplit=1)[0]
        desc = description.replace("`", "") if description else ""
        # some docstring params may be mislabeled and not appear in the params
        if argname in sig.parameters:
            # use setdefault so as not to override an explicitly provided tooltip
            sig.parameters[argname].options.setdefault("tooltip", desc)


_R = TypeVar("_R")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class FunctionGui(Container, Generic[_P, _R]):
    """Wrapper for a container of widgets representing a callable object.

    Parameters
    ----------
    function : Callable
        A callable to turn into a GUI
    call_button : bool | str | None, optional
        If True, create an additional button that calls the original function when
        clicked.  If a `str`, set the button text. by default False when
        auto_call is True, and True otherwise.
        The button can be accessed from the `.call_button` property.
    layout : str, optional
        The type of layout to use. Must be `horizontal` or `vertical`
        by default "horizontal".
    scrollable : bool, optional
        Whether to enable scroll bars or not. If enabled, scroll bars will
        only appear along the layout direction, not in both directions.
    labels : bool, optional
        Whether labels are shown in the widget. by default True
    tooltips : bool, optional
        Whether tooltips are shown when hovering over widgets. by default True
    app : Application | str | None, optional
        A backend to use, by default `None` (use the default backend.)
    visible : bool, optional
        Whether to immediately show the widget.  If `False`, widget is explicitly
        hidden.  If `None`, widget is not shown, but will be shown if a parent
        container is shown, by default None.
    auto_call : bool, optional
        If True, changing any parameter in either the GUI or the widget attributes
        will call the original function with the current settings. by default False
    result_widget : bool, optional
        Whether to display a LineEdit widget the output of the function when called,
        by default False
    param_options : dict, optional
        A dict of name: widget_options dict for each parameter in the function.
        Will be passed to `magic_signature` by default `None`
    name : str, optional
        A name to assign to the Container widget, by default `function.__name__`
    persist : bool, optional
        If `True`, when parameter values change in the widget, they will be stored to
        disk (in `~/.config/magicgui/cache`) and restored when the widget is loaded
        again with `persist = True`.  By default, `False`.
    raise_on_unknown : bool
        If True, raise an error if a parameter annotation is not recognized.

    Raises
    ------
    TypeError
        If unexpected keyword arguments are provided
    """

    called = Signal(
        object, description="Emitted with the result after the function is called."
    )
    _widget: ContainerProtocol

    def __init__(
        self,
        function: Callable[_P, _R],
        call_button: bool | str | None = None,
        layout: str = "vertical",
        scrollable: bool = False,
        labels: bool = True,
        tooltips: bool = True,
        app: AppRef | None = None,
        visible: bool | None = None,
        auto_call: bool = False,
        result_widget: bool = False,
        param_options: dict[str, dict] | None = None,
        name: str | None = None,
        persist: bool = False,
        raise_on_unknown: bool = False,
        **kwargs: Any,
    ):
        if not callable(function):
            raise TypeError("'function' argument to FunctionGui must be callable.")

        # consume extra Widget keywords
        extra = set(kwargs) - {"annotation", "gui_only"}
        if extra:
            s = "s" if len(extra) > 1 else ""
            raise TypeError(f"FunctionGui got unexpected keyword argument{s}: {extra}")
        if param_options is None:
            param_options = {}
        elif not isinstance(param_options, dict):
            raise TypeError("'param_options' must be a dict of dicts")

        sig = magic_signature(
            function, gui_options=param_options, raise_on_unknown=raise_on_unknown
        )
        self._return_annotation = resolve_single_type(sig.return_annotation)
        self._tooltips = tooltips
        if tooltips:
            _inject_tooltips_from_docstrings(function.__doc__, sig)

        self.persist = persist
        self._function = function
        self.__wrapped__ = function
        # it's conceivable that function is not actually an instance of FunctionType
        # we can still support any generic callable, but we need to be careful not to
        # access attributes (like `__name__` that only function objects have).
        # Mypy doesn't seem catch this at this point:
        # https://github.com/python/mypy/issues/9934
        name = name or getattr(function, "__name__", None)
        if not name:
            if hasattr(function, "func"):  # partials:
                f = getattr(function, "func", None)
                name = getattr(f, "__name__", None) or str(function)
            else:
                name = f"{function.__module__}.{function.__class__}"
        self._callable_name = name

        super().__init__(
            layout=layout,
            scrollable=scrollable,
            labels=labels,
            visible=visible,
            widgets=list(sig.widgets(app).values()),
            name=name or self._callable_name,
        )
        self._param_options = param_options
        self._result_name = ""
        self._call_count: int = 0
        self._bound_instances: dict[int, FunctionGui] = {}

        # a deque of Progressbars to be created by (possibly nested) tqdm_mgui iterators
        self._tqdm_pbars: deque[ProgressBar] = deque()
        # the nesting level of tqdm_mgui iterators in a given __call__
        self._tqdm_depth: int = 0

        if call_button is None:
            call_button = not auto_call
        self._call_button: PushButton | None = None
        if call_button:
            text = call_button if isinstance(call_button, str) else "Run"
            self._call_button = PushButton(gui_only=True, text=text, name="call_button")
            if not auto_call:  # (otherwise it already gets called)

                @self._call_button.changed.connect
                def _disable_button_and_call() -> None:
                    # disable the call button until the function has finished
                    self._call_button = cast(PushButton, self._call_button)
                    self._call_button.enabled = False
                    try:
                        self.__call__()
                    finally:
                        self._call_button.enabled = True

            self.append(self._call_button)

        self._result_widget: ValueWidget | None = None
        if result_widget:
            from magicgui.widgets.bases import create_widget

            self._result_widget = cast(
                ValueWidget,
                create_widget(
                    value=None,
                    annotation=self._return_annotation,
                    gui_only=True,
                    is_result=True,
                    raise_on_unknown=raise_on_unknown,
                ),
            )
            self.append(self._result_widget)

        if persist:
            self._load(quiet=True)

        self._auto_call = auto_call
        self.changed.connect(self._on_change)

    def _on_change(self) -> None:
        if self.persist:
            self._dump()
        if self._auto_call:
            self()

    @property
    def call_button(self) -> PushButton | None:
        """Return the call button."""
        return self._call_button

    @property
    def call_count(self) -> int:
        """Return the number of times the function has been called."""
        return self._call_count

    def reset_call_count(self) -> None:
        """Reset the call count to 0."""
        self._call_count = 0

    @property
    def return_annotation(self) -> Any:
        """Return annotation for [inspect.Signature][inspect.Signature] conversion.

        ForwardRefs will be resolve when setting the annotation.
        """
        return self._return_annotation

    @property
    def __signature__(self) -> MagicSignature:
        """Return a MagicSignature object representing the current state of the gui."""
        return super().__signature__.replace(return_annotation=self.return_annotation)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        """Call the original function with the current parameter values from the Gui.

        You may pass a `update_widget=True` keyword argument to update the widget
        values to match the current parameter values before calling the function.

        It is also possible to override the current parameter values from the GUI by
        providing args/kwargs to the function call.  Only those provided will override
        the ones from the gui.  A `called` signal will also be emitted with the results.

        Returns
        -------
        result : Any
            whatever the return value of the original function would have been.

        Examples
        --------
        ```python
        gui = FunctionGui(func, show=True)

        # then change parameters in the gui, or by setting:  gui.param.value = something

        gui()  # calls the original function with the current parameters
        ```
        """
        update_widget: bool = bool(kwargs.pop("update_widget", False))

        sig = self.__signature__
        try:
            bound = sig.bind(*args, **kwargs)
        except TypeError as e:
            if "missing a required argument" not in str(e):
                raise

            match = re.search("argument: '(.+)'", str(e))
            missing = match.groups()[0] if match else "<param>"
            msg = (
                f"{e} in call to '{self._callable_name}{sig}'.\n"
                "To avoid this error, you can bind a value or callback to the "
                f"parameter:\n\n    {self._callable_name}.{missing}.bind(value)"
                "\n\nOr use the 'bind' option in the magicgui decorator:\n\n"
                f"    @magicgui({missing}={{'bind': value}})\n"
                f"    def {self._callable_name}{sig}: ..."
            )
            raise TypeError(msg) from None

        if update_widget:
            self._auto_call, before = False, self._auto_call
            try:
                self.update(bound.arguments)
            finally:
                self._auto_call = before

        bound.apply_defaults()

        self._tqdm_depth = 0  # reset the tqdm stack count
        with _function_name_pointing_to_widget(self):
            value = self._function(*bound.args, **bound.kwargs)

        self._call_count += 1
        if self._result_widget is not None:
            with self._result_widget.changed.blocked():
                self._result_widget.value = value

        return_type = sig.return_annotation
        if return_type:
            from magicgui.type_map import type2callback

            for callback in type2callback(return_type):
                callback(self, value, return_type)
        self.called.emit(value)
        return value

    def __repr__(self) -> str:
        """Return string representation of instance."""
        return f"<{type(self).__name__} {self._callable_name}{self.__signature__}>"

    @property
    def result_name(self) -> str:
        """Return a name that can be used for the result of this magicfunction."""
        return self._result_name or f"{self._callable_name} result"

    @result_name.setter
    def result_name(self, value: str) -> None:
        """Set the result name of this FunctionGui widget."""
        self._result_name = value

    def copy(self) -> FunctionGui:
        """Return a copy of this FunctionGui."""
        return FunctionGui(
            function=self._function,
            call_button=self._call_button.text if self._call_button else None,
            layout=self.layout,
            labels=self.labels,
            param_options=self._param_options,
            auto_call=self._auto_call,
            result_widget=bool(self._result_widget),
            app=None,
            persist=self.persist,
            visible=self.visible,
            tooltips=self._tooltips,
            scrollable=self._scrollable,
            name=self.name,
        )

    def __get__(self, obj: object, objtype: type | None = None) -> FunctionGui:
        """Provide descriptor protocol.

        This allows the @magicgui decorator to work on a function as well as a method.
        If a method on a class is decorated with `@magicgui`, then accessing the
        attribute on an instance of that class will return a version of the FunctionGui
        in which the first argument of the function is bound to the instance. (Just like
        what you'd expect with the @property decorator.)

        Returns
        -------
        bound : FunctionGui
            A new FunctionGui instance.

        Examples
        --------
        ```python
        >>> class MyClass:
        ...     @magicgui
        ...     def my_method(self, x=1):
        ...         print(locals())
        ...
        >>> c = MyClass()
        >>> c.my_method  # the FunctionGui that can be used as a widget

        # calling it works as usual, with `c` provided as `self`
        >>> c.my_method(x=34)
        {'self': <__main__.MyClass object at 0x7fb610e455e0>, 'x': 34}
        ```
        """
        if obj is None:
            return self
        obj_id = id(obj)
        if obj_id not in self._bound_instances:
            method = getattr(obj.__class__, self._function.__name__)
            p0 = next(iter(inspect.signature(method).parameters))
            prior, self._param_options = (
                self._param_options,
                {
                    p0: {"bind": obj},
                    **self._param_options,
                },
            )
            try:
                self._bound_instances[obj_id] = self.copy()
            finally:
                self._param_options = prior
        return self._bound_instances[obj_id]

    def __set__(self, obj: Any, value: Any) -> NoReturn:
        """Prevent setting a magicgui attribute."""
        raise AttributeError("Can't set magicgui attribute")

    @property
    def _dump_path(self) -> Path:
        from magicgui._util import user_cache_dir

        name = getattr(self._function, "__qualname__", self._callable_name)
        name = name.replace("<", "-").replace(">", "-")  # e.g. <locals>
        return user_cache_dir() / f"{self._function.__module__}.{name}"

    def _dump(self, path: str | Path | None = None) -> None:
        super()._dump(path or self._dump_path)

    def _load(self, path: str | Path | None = None, quiet: bool = False) -> None:
        super()._load(path or self._dump_path, quiet=quiet)


class MainFunctionGui(FunctionGui[_P, _R], MainWindow):
    """Container of widgets as a Main Application Window."""

    _widget: MainWindowProtocol

    def __init__(self, function: Callable[_P, _R], *args: Any, **kwargs: Any) -> None:
        super().__init__(function, *args, **kwargs)
        self.create_menu_item("Help", "Documentation", callback=self._show_docs)
        self._help_text_edit: TextEdit | None = None

    def _show_docs(self) -> None:
        if not self._help_text_edit:
            from magicgui.widgets import TextEdit

            docs = self._function.__doc__
            html = _docstring_to_html(docs) if docs else "None"
            self._help_text_edit = TextEdit(value=html)
            self._help_text_edit.read_only = True
            self._help_text_edit.width = 600
            self._help_text_edit.height = 400

        self._help_text_edit.show()


def _docstring_to_html(docs: str) -> str:
    """Convert docstring into rich text html."""
    from docstring_parser import parse

    ds = parse(docs)

    ptemp = "<li><p><strong>{}</strong> (<em>{}</em>) - {}</p></li>"
    plist = [ptemp.format(p.arg_name, p.type_name, p.description) for p in ds.params]
    params = f'<h3>Parameters</h3><ul>{"".join(plist)}</ul>'
    short = f"<p>{ds.short_description}</p>" if ds.short_description else ""
    long = f"<p>{ds.long_description}</p>" if ds.long_description else ""
    return re.sub(r"`?([^`]+)`?", r"<code>\1</code>", f"{short}{long}{params}")


@contextmanager
def _function_name_pointing_to_widget(function_gui: FunctionGui) -> Iterator[None]:
    """Context in which the name of the function points to the function_gui instance.

    When calling the function provided to FunctionGui, we make sure that the name
    of the function points to the FunctionGui object itself.
    In standard `@magicgui` usage, this will have been the case anyway.
    Doing this here allows the function name in a `@magic_factory`-decorated function
    to *also* refer to the function gui instance created by the factory, (rather than
    to the [~magicgui._magicgui.MagicFactory][magicgui._magicgui.MagicFactory] object).

    Examples
    --------
    >>> @magicgui
    >>> def func():
    ...     # using "func" in the body here will refer to the widget.
    ...     print(type(func))
    >>>
    >>> func()  # prints 'magicgui.widgets._function_gui.FunctionGui'

    >>> @magic_factory
    >>> def func():
    ...     # using "func" in the body here will refer to the *widget* not the factory.
    ...     print(type(func))
    >>>
    >>> widget = func()
    >>> widget()  # *also* prints 'magicgui.widgets._function_gui.FunctionGui'
    """
    function = function_gui._function
    if not isinstance(function, FunctionType):
        # it's not a function object, so we don't know how to patch it...
        yield
        return

    func_name = function.__name__
    # see https://docs.python.org/3/library/inspect.html for details on code objects
    code = function.__code__

    if func_name in code.co_names:
        # This indicates that the function name was used inside the body of the
        # function, and points to some object in the module's global namespace.
        # function.__globals__ here points to the module-level globals in which the
        # function was defined.
        original_value = function.__globals__.get(func_name)
        function.__globals__[func_name] = function_gui
        try:
            yield
        finally:
            function.__globals__[func_name] = original_value

    elif function.__closure__ and func_name in code.co_freevars:
        # This indicates that the function name was used inside the body of the
        # function, and points to some object defined in a local scope (closure), rather
        # than the module's global namespace.
        # the position of the function name in code.co_freevars tells us where to look
        # for the value in the function.__closure__ tuple.
        idx = code.co_freevars.index(func_name)
        original_value = function.__closure__[idx].cell_contents
        function.__closure__[idx].cell_contents = function_gui
        try:
            yield
        finally:
            function.__closure__[idx].cell_contents = original_value
    else:
        yield



"""# Range slider widget

A double ended range slider widget.
"""

from typing import Tuple

from magicgui import magicgui


@magicgui(auto_call=True, range_value={"widget_type": "RangeSlider", "max": 500})
def func(range_value: Tuple[int, int] = (20, 380)):
    """Double ended range slider."""
    print(range_value)


func.show(run=True)



"""# napari parameter sweeps

[napari](https://github.com/napari/napari) is a fast, interactive,
multi-dimensional image viewer for python.  It uses Qt for the GUI, so it's easy
to extend napari with small, composable widgets created with `magicgui`.  Here,
we demonstrate how to build a interactive widget that lets you immediately see
the effect of changing one of the parameters of your function.

For napari-specific magicgui documentation, see the
[napari docs](https://napari.org/guides/magicgui.html)

![napari parameter sweep widget](../../images/param_sweep.gif){ width=80% }

*See also:* Some of this tutorial overlaps with topics covered in the
[napari image arithmetic example](napari_img_math.py).

## outline

**This example demonstrates how to:**

1. Create a `magicgui` widget that can be used in another
program (`napari`)

2. Automatically call our function when a parameter changes

3. Provide `magicgui` with a custom widget for a specific
argument

4. Use the `choices` option to create a dropdown

5. Connect some event listeners to create interactivity.

## code

*Code follows, with explanation below... You can also [get this example at
github](https://github.com/pyapp-kit/magicgui/blob/main/docs/examples/napari/napari_param_sweep.py).*

```python linenums="1" hl_lines="14-19 31"
import napari
import skimage.data
import skimage.filters
from napari.types import ImageData

from magicgui import magicgui


# turn the gaussian blur function into a magicgui
# - 'auto_call' tells magicgui to call the function when a parameter changes
# - we use 'widget_type' to override the default "float" widget on sigma,
#   and provide a maximum valid value.
# - we contstrain the possible choices for 'mode'
@magicgui(
    auto_call=True,
    sigma={"widget_type": "FloatSlider", "max": 6},
    mode={"choices": ["reflect", "constant", "nearest", "mirror", "wrap"]},
    layout="horizontal",
)
def gaussian_blur(layer: ImageData, sigma: float = 1.0, mode="nearest") -> ImageData:
    # Apply a gaussian blur to 'layer'.
    if layer is not None:
        return skimage.filters.gaussian(layer, sigma=sigma, mode=mode)

# create a viewer and add some images
viewer = napari.Viewer()
viewer.add_image(skimage.data.astronaut().mean(-1), name="astronaut")
viewer.add_image(skimage.data.grass().astype("float"), name="grass")

# Add it to the napari viewer
viewer.window.add_dock_widget(gaussian_blur)
# update the layer dropdown menu when the layer list changes
viewer.layers.events.changed.connect(gaussian_blur.reset_choices)

napari.run()
```

## walkthrough

We're going to go a little out of order so that the other code makes more sense.
Let's start with the actual function we'd like to write to apply a gaussian
filter to an image.

### the function

Our function is a very thin wrapper around
[`skimage.filters.gaussian`](https://scikit-image.org/docs/dev/api/skimage.filters.html#skimage.filters.gaussian).
It takes a `napari` [Image
layer](https://napari.org/howtos/layers/image.html), a `sigma` to control
the blur radius, and a `mode` that determines how edges are handled.

```python
def gaussian_blur(
    layer: Image, sigma: float = 1, mode="nearest"
) -> Image:
    return filters.gaussian(layer.data, sigma=sigma, mode=mode)
```

The reasons we are wrapping it here are:

1. `filters.gaussian` accepts a `numpy` array,
   but we want to work with `napari` layers
   that store the data in a `layer.data` attribute. So we need an adapter.
2. We'd like to add some [type annotations](../../type_map.md) to the
   signature that were not provided by `filters.gaussian`

#### type annotations

As described in the [image arithmetic example](napari_img_math.py), we take
advantage of napari's built in support for `magicgui` by annotating our function
parameters and return value as napari `Layer` types.  `napari` will then tell
`magicgui` what to do with them, creating a dropdown with a list of current
layers for our `layer` parameter, and automatically adding the result of our
function to the viewer when called.

For documentation on napari types with magicgui, see the
[napari docs](https://napari.org/guides/magicgui.html)

### the magic part

Finally, we decorate the function with `@magicgui` and provide some options.

```python
@magicgui(
    auto_call=True,
    sigma={"widget_type": "FloatSlider", "max": 6},
    mode={"choices": ["reflect", "constant", "nearest", "mirror", "wrap"]},
)
def gaussian_blur(
  layer: ImageData, sigma: float = 1.0, mode="nearest"
) -> ImageData:
    # Apply a gaussian blur to ``layer``.
    if layer is not None:
        return skimage.filters.gaussian(layer, sigma=sigma, mode=mode)
```

- `auto_call=True` makes it so that the `gaussian_blur` function will be called
  whenever one of the parameters changes (with the current parameters set in the
  GUI).
- We then provide keyword arguments to modify the look & behavior of `sigma`
  and `mode`:

    - `"widget_type": "FloatSlider"` tells `magicgui` not to use the standard
        (`float`) widget for the `sigma` widget, but rather to use a slider widget.
    - we then set an upper limit on the slider values for `sigma`.

- finally, we specify valid `choices` for the `mode` argument.  This turns that
  parameter into a categorical/dropdown type widget, and sets the options.

### connecting events

As described in the [Events documentation](../../events.md), we can
also connect any callback to the `gaussian_blur.called` signal that will receive
the result of our decorated function anytime it is called.

```python
def do_something_with_result(result):
    ...

gaussian_blur.called.connect(do_something_with_result)
```

## Code

Here's the full code example again.
"""

# %%
import napari
import skimage.data
import skimage.filters
from napari.types import ImageData

from magicgui import magicgui


# turn the gaussian blur function into a magicgui
# - 'auto_call' tells magicgui to call the function when a parameter changes
# - we use 'widget_type' to override the default "float" widget on sigma,
#   and provide a maximum valid value.
# - we contstrain the possible choices for 'mode'
@magicgui(
    auto_call=True,
    sigma={"widget_type": "FloatSlider", "max": 6},
    mode={"choices": ["reflect", "constant", "nearest", "mirror", "wrap"]},
    layout="horizontal",
)
def gaussian_blur(layer: ImageData, sigma: float = 1.0, mode="nearest") -> ImageData:
    # Apply a gaussian blur to 'layer'.
    if layer is not None:
        return skimage.filters.gaussian(layer, sigma=sigma, mode=mode)


# create a viewer and add some images
viewer = napari.Viewer()
viewer.add_image(skimage.data.astronaut().mean(-1), name="astronaut")
viewer.add_image(skimage.data.grass().astype("float"), name="grass")

# Add it to the napari viewer
viewer.window.add_dock_widget(gaussian_blur)
# update the layer dropdown menu when the layer list changes
viewer.layers.events.changed.connect(gaussian_blur.reset_choices)

napari.run()



"""# Basic widgets demo

Widget demonstration with magicgui.

This code demonstrates a few of the widget types that magicgui can create
based on the parameter types in your function.
"""

import datetime
from enum import Enum
from pathlib import Path

from magicgui import magicgui


class Medium(Enum):
    """Enum for various media and their refractive indices."""

    Glass = 1.520
    Oil = 1.515
    Water = 1.333
    Air = 1.0003


@magicgui(
    main_window=True,
    call_button="Calculate",
    layout="vertical",
    result_widget=True,
    slider_float={"widget_type": "FloatSlider", "max": 100},
    slider_int={"widget_type": "Slider", "readout": False},
    radio_option={
        "widget_type": "RadioButtons",
        "orientation": "horizontal",
        "choices": [("first option", 1), ("second option", 2)],
    },
    filename={"label": "Pick a file:"},
)
def widget_demo(
    boolean=True,
    integer=1,
    spin_float=3.14159,
    slider_float=43.5,
    slider_int=550,
    string="Text goes here",
    dropdown=Medium.Glass,
    radio_option=2,
    date=datetime.date(1999, 12, 31),
    time=datetime.time(1, 30, 20),
    datetime=datetime.datetime.now(),
    filename=Path.home(),
):
    """We can use numpy docstrings to provide tooltips.

    Parameters
    ----------
    boolean : bool, optional
        A checkbox for booleans, by default True
    integer : int, optional
        Some integer, by default 1
    spin_float : float, optional
        This one is a float, by default "pi"
    slider_float : float, optional
        Hey look! I'm a slider, by default 43.5
    slider_int : float, optional
        I only take integers, and I've hidden my readout, by default 550
    string : str, optional
        We'll use this string carefully, by default "Text goes here"
    dropdown : Enum, optional
        Pick a medium, by default Medium.Glass
    radio_option : int
        A set of radio buttons.
    date : datetime.date, optional
        Your birthday, by default datetime.date(1999, 12, 31)
    time : datetime.time, optional
        Some time, by default datetime.time(1, 30, 20)
    datetime : datetime.datetime, optional
        A very specific time and date, by default ``datetime.datetime.now()``
    filename : str, optional
        Pick a path, by default Path.home()
    """
    return locals().values()


widget_demo.changed.connect(print)
widget_demo.show(run=True)



import re
import runpy
import sys
from glob import glob
from pathlib import Path

import pytest

from magicgui import type_map, use_app


@pytest.mark.parametrize(
    "fname",
    [
        f
        for f in glob("docs/**/*.md", recursive=True)
        if "_build" not in f and Path(f).read_text(encoding="utf-8").startswith("-")
    ],
)
def test_doc_code_cells(fname):
    """Make sure that all code cells in documentation perform as expected."""
    globalns = globals()
    text = Path(fname).read_text()
    code_cells = re.findall(r"```{code-cell}[^\n]+\n(.*?)`{3}", text, re.S)
    for cell in code_cells:
        header = re.search(r"-{3}(.+?)-{3}", cell, re.S)
        if header:
            cell = cell.replace(header.group(), "")
            if "warns" in header.group():
                with pytest.warns(None):
                    exec(cell, globalns)
                continue
        exec(cell, globalns)


EXAMPLES = Path(__file__).parent.parent / "docs" / "examples"
# leaving out image only because finding the image file in both
# tests and docs is a pain...
# range_slider has periodic segfaults
EXCLUDED = {"napari", "image", "range_slider"}
example_files = [
    str(f) for f in EXAMPLES.rglob("*.py") if all(x not in str(f) for x in EXCLUDED)
]

# if os is Linux and python version is 3.9 and backend is PyQt5
LINUX = sys.platform.startswith("linux")
PY39 = sys.version_info >= (3, 9)
PYQT5 = "PyQt5" in sys.modules
if LINUX and PY39 and PYQT5:
    # skip range_slider example because of superqt c++ wrapped item bug
    example_files = [f for f in example_files if "range_slider" not in f]


@pytest.mark.parametrize("fname", example_files)
def test_examples(fname, monkeypatch):
    """Make sure that all code cells in documentation perform as expected."""
    if "values_dialog" in str(fname):
        from magicgui.backends._qtpy.widgets import QtW  # type: ignore

        try:
            monkeypatch.setattr(QtW.QDialog, "exec", lambda s: None)
        except AttributeError:
            monkeypatch.setattr(QtW.QDialog, "exec_", lambda s: None)
    app = use_app()
    app.start_timer(50 if "table" in str(fname) else 5, app.quit)
    try:
        runpy.run_path(fname)
    except ImportError as e:
        if "Numpy required to use images" in str(e):
            pytest.skip("numpy unavailable: skipping image example")
    finally:
        if "waveform" in fname:
            type_map._type_map._TYPE_DEFS.pop(int, None)
            type_map._type_map._TYPE_DEFS.pop(float, None)



from __future__ import annotations

import operator
import sys
from itertools import zip_longest
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Generic,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Literal,
    Mapping,
    MutableMapping,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)
from warnings import warn

from magicgui.application import use_app
from magicgui.widgets.bases._mixins import _ReadOnlyMixin
from magicgui.widgets.bases._value_widget import ValueWidget

if TYPE_CHECKING:
    import numpy
    import pandas
    from typing_extensions import TypeGuard, Unpack

    from magicgui.widgets.protocols import TableWidgetProtocol

    from .bases._widget import WidgetKwargs

TblKey = Any
_KT = TypeVar("_KT")  # Key type
_KT_co = TypeVar("_KT_co", covariant=True)  # Key type covariant containers.
_VT_co = TypeVar("_VT_co", covariant=True)  # Value type covariant containers.
TableData = Union[dict, "pandas.DataFrame", list, "numpy.ndarray", tuple, None]
IndexKey = Union[int, slice]
SliceNone = slice(None)


def normalize_table_data(data: TableData) -> tuple[Collection[Collection], list, list]:
    """Convert data to data, row headers, column headers.

    Parameters
    ----------
    data : dict, dataframe, list, array, tuple, optional
        Table data (and/or header data), in one of the accepted formats:

        - list or list-of-lists : [column_values] or [[row_vals], ..., [row_vals]]
        - dict-of-dicts : {column_header -> {row_header -> value}}
        - dict-of-lists : {column_header -> [column_values]}
        - list-of-row-records :
            [{column_headers -> value}, ... , {column_headers -> value}]
        - split-dict-of-lists :
            {'data' -> [values], 'index' -> [index], 'columns' -> [columns]}
        - tuple-of-values : ([values], [row_headers], [column_headers])
        - dict-of-pandas-series : {column_header -> Series(values)}

    """
    if data is None:
        return [], [], []
    if isinstance(data, dict):
        return _from_dict(data)
    if isinstance(data, tuple):
        data_len = len(data)
        _data = data[0] if data else []
        _index = data[1] if data_len > 1 else []
        _columns = data[2] if data_len > 2 else []
        return _data, _index, _columns
    if _is_dataframe(data):
        return data.values, data.index, data.columns
    if isinstance(data, list):
        if data:
            if isinstance(data[0], dict):
                return _from_records(data)
            if not isinstance(data[0], Collection):
                # single column dataset
                return [[i] for i in data], [], []
        return data, [], []
    if _is_numpy_array(data):
        return data, [], []
    raise TypeError(
        f"Table value must be a dict, dataframe, list, or array, got {type(data)}"
    )


class HeadersView(KeysView[_KT]):
    """dictionary view for Table headers."""

    def __init__(self, mapping: Table, axis: str = "column") -> None:
        super().__init__(mapping)
        self._mapping = mapping
        axis = axis.rstrip("s")
        assert axis in {"row", "column"}, "keys axis must be either 'column' or 'row'"
        self._axis = axis

    def __iter__(self) -> Iterator[_KT_co]:
        """Yield headers."""
        yield from getattr(self._mapping, f"{self._axis}_headers")

    def __repr__(self) -> str:
        """Return string repr of column headers view."""
        return f"{self._axis}_headers({list(self)})"


class TableItemsView(ItemsView[_KT_co, _VT_co], Generic[_KT_co, _VT_co]):
    """dictionary view for Table items."""

    def __init__(self, mapping: Mapping[_KT_co, _VT_co], axis: str = "column") -> None:
        super().__init__(mapping)
        self._mapping: Table = mapping  # type: ignore
        axis = axis.rstrip("s")
        assert axis in {"row", "column"}, "keys axis must be either 'column' or 'row'"
        self._axis = axis

    def __iter__(self) -> Iterator[tuple[_KT_co, _VT_co]]:
        """Yield items."""
        for header in getattr(self._mapping, f"{self._axis}_headers"):
            val = getattr(self._mapping, f"_get_{self._axis}")(header)
            yield (header, val)

    def __repr__(self) -> str:
        """Return string repr of column headers view."""
        n = self._mapping.shape[0 if self._axis == "row" else 1]
        return f"table_items({n} {self._axis}s)"


class Table(ValueWidget, _ReadOnlyMixin, MutableMapping[TblKey, list]):
    """A widget to represent columnar or 2D data with headers.

    Tables behave like plain `dicts`, where the keys are column headers and the
    (list-like) values are column data.

    Parameters
    ----------
    value : dict, dataframe, list, array, tuple, optional
        Table data (and/or header data), in one of the accepted formats:

        - list or list-of-lists : [column_values] or [[row_vals], ..., [row_vals]]
        - dict-of-dicts : {column_header -> {row_header -> value}}
        - dict-of-lists : {column_header -> [column_values]}
        - list-of-row-records :
            [{column_headers -> value}, ... , {column_headers -> value}]
        - split-dict-of-lists :
            {'data' -> [values], 'index' -> [index], 'columns' -> [columns]}
        - tuple-of-values : ([values], [row_headers], [column_headers])
        - dict-of-pandas-series : {column_header -> Series(values)}

    index : Collection, optional
        A sized iterable container of row headers. By default, row headers will be
        ``tuple(range(len(data)))``.  Values provided here override any implied in
        ``value``.
    columns : Collection, optional
        A sized iterable container of column headers. By default, column headers will be
        ``tuple(range(len(data[0])))``.  Values provided here override any implied in
        ``value``.
    **kwargs
        Additional kwargs will be passed to the
        [magicgui.widgets.Widget][magicgui.widgets.Widget] constructor.

    Attributes
    ----------
    value : dict
        Returns a dict with the keys `data`, `index`, and `columns` ... representing the
        2D (list of lists) tabular data, row headers, and column headers, respectively.
        If set, will clear and update the table using the new data.
    data : DataView
        A `DataView` instance that provides numpy-like indexing (with
        get/set/delete) onto the 2D data array,  For example `table.data[0,2]` gets the
        data in the cell of the first row, 3rd column.  Works with numpy slice syntax.
    column_headers : tuple
        The current column headers.  Can be set with a new sequence to change
    row_headers : tuple
        The current row headers.  Can be set with a new sequence to change
    shape : tuple of int
        The shape of the table in `(rows, columns)`.
    size : int
        The number of cells in the table.

    Methods
    -------
    keys(axis='column')
        Return a `TableHeadersView`,
        providing a view on this table's headers. Use `axis='row'` for row headers.
    items(axis='column')
        Return a `TableItemsView`,
        providing a view on this table's items, as 2-tuples of `(header, data)`. Use
        `axis='row'` for `(row_header, row_data)`
    clear()
        Clear all table data and headers.
    to_dataframe()
        Returns a pandas dataframe representation of this table. (requires pandas)
    to_dict(orient='dict')
        Return one of many different dict-like representations of table and header data.
        See docstring of :meth:`to_dict` for details.

    Events
    ------
    changed
        Emitted whenever a cell in the table changes. The value will have a
        dict of information regarding the cell that changed:
        {'data': x, 'row': int, 'column': int, 'column_header': str, 'row_header': str}
        CURRENTLY: only emitted on changes in the GUI. not programmatic changes.
    """

    _widget: TableWidgetProtocol

    def __new__(
        cls,
        value: TableData | None = None,
        *,
        index: Collection | None = None,
        columns: Collection | None = None,
        **kwargs: Any,
    ) -> Table:
        """Just for the signature."""
        return super().__new__(cls)

    def __init__(
        self,
        value: TableData | None = None,
        *,
        index: Collection | None = None,
        columns: Collection | None = None,
        **kwargs: Unpack[WidgetKwargs],
    ) -> None:
        kwargs["widget_type"] = use_app().get_obj("Table")
        super().__init__(**kwargs)
        self._data = DataView(self)
        data, _index, _columns = normalize_table_data(value)
        self.value = {
            "data": data,
            "index": index if index is not None else _index,
            "columns": columns if columns is not None else _columns,
        }

    @property
    def value(self) -> dict[TblKey, Collection]:
        """Return dict with current `data`, `index`, and `columns` of the widget."""
        return self.to_dict("split")

    @value.setter
    def value(self, value: TableData) -> None:
        """Set table data from dict, dataframe, list, or array.

        Parameters
        ----------
        value : Any
            Complete table data in one of the forms described above. Partial table
            updates are not yet supported
        """
        data, index, columns = normalize_table_data(value)
        _validate_table_data(data, index, columns)
        self.clear()
        try:
            nc = len(data[0])  # type: ignore
        except (TypeError, IndexError):
            nc = 0
        self.column_headers = tuple(columns) or range(nc)  # type:ignore
        self.row_headers = tuple(index) or range(len(data))  # type: ignore
        for row, d in enumerate(data):
            self._set_rowi(row, d)

    def delete_row(
        self,
        *,
        index: int | Sequence[int] | None = None,
        header: Any | Sequence[Any] | None = None,
    ) -> None:
        """Delete row(s) by index or header.

        Parameters
        ----------
        index : int or Sequence[int], optional
            Index or indices of row(s) to delete.
        header : Any or Sequence[Any], optional
            Header or headers of row(s) to delete.
        """
        indices: set[int] = set()
        if index is not None:
            if isinstance(index, Sequence):
                indices.update(index)
            else:
                indices.add(index)
        if header is not None:
            if isinstance(header, str) or not isinstance(header, Sequence):
                header = (header,)
            row_headers = self.row_headers
            for h in header:
                try:
                    indices.add(row_headers.index(h))
                except ValueError as e:
                    raise KeyError(f"{h!r} is not a valid row header") from e
        for i in sorted(indices, reverse=True):
            self._del_rowi(i)

    @property
    def data(self) -> DataView:
        """Return DataView object for this table."""
        return self._data

    @data.setter
    def data(self, value: TableData) -> None:
        """Set 2D table data."""
        self._data.__setitem__(slice(None), value)

    @property
    def column_headers(self) -> tuple:
        """Return column headers."""
        headers = self._widget._mgui_get_column_headers()
        if not headers:
            ncols = self._widget._mgui_get_column_count()
            headers = tuple(range(ncols))
        return headers

    @column_headers.setter
    def column_headers(self, headers: Sequence) -> None:
        """Set column headers."""
        self._check_new_headers(headers, axis="column")
        return self._widget._mgui_set_column_headers(headers)

    @property
    def row_headers(self) -> tuple:
        """Return row headers."""
        nrows = self._widget._mgui_get_row_count()
        return self._widget._mgui_get_row_headers() or tuple(range(nrows))

    @row_headers.setter
    def row_headers(self, headers: Sequence) -> None:
        """Set row headers."""
        self._check_new_headers(headers, axis="row")
        return self._widget._mgui_set_row_headers(headers)

    @property
    def shape(self) -> tuple[int, int]:
        """Return shape of table widget (rows, cols)."""
        return self._widget._mgui_get_row_count(), self._widget._mgui_get_column_count()

    # # Should we allow this?
    # @shape.setter
    # def shape(self, shape: Tuple[int, int]) -> None:
    #     """Set shape of table widget (rows, cols)."""
    #     try:
    #         r, c = shape[:2]
    #         r = int(r)
    #         c = int(c)
    #     except (ValueError, TypeError):
    #         raise ValueError("'shape' argument must be an iterable of 2 integers")
    #     self._widget._mgui_set_row_count(r)
    #     self._widget._mgui_set_column_count(c)
    #     # TODO: need to truncate extend headers as necessary

    @property
    def size(self) -> int:
        """Return shape of table widget (rows, cols)."""
        return cast(int, operator.mul(*self.shape))

    def keys(self, axis: str = "column") -> HeadersView[TblKey]:
        """Return a set-like object providing a view on this table's headers."""
        return HeadersView(self, axis)

    def items(self, axis: str = "column") -> TableItemsView[TblKey, list]:
        """Return a set-like object providing a view on this table's items."""
        return TableItemsView(self, axis)

    def clear(self) -> None:
        """Clear the table."""
        self._widget._mgui_set_row_count(0)
        self._widget._mgui_set_column_count(0)

    def __delitem__(self, key: TblKey) -> None:
        """Delete a column from the table."""
        self._del_column(key)

    def __getitem__(self, key: TblKey) -> list:
        """Get a column from the table."""
        return self._get_column(key)

    def __setitem__(self, key: TblKey, v: Collection) -> None:
        """Set a column in the table. If `k` doesn't exist, make a new column."""
        self._set_column(key, v)

    def __iter__(self) -> Iterator:
        """Yield column headers."""
        yield from self.column_headers

    def __len__(self) -> int:
        """Return number of columns."""
        return self._widget._mgui_get_column_count()

    def __hash__(self) -> int:
        """Make table hashable."""
        return id(self)

    def __repr__(self) -> str:
        """Return string repr."""
        name = f"name={self.name!r}, " if self.name else ""
        return f"Table({name}shape={self.shape} at {hex(id(self))})"

    def _check_new_headers(self, headers: Sequence, *, axis: str = "column") -> None:
        current_headers = getattr(self._widget, f"_mgui_get_{axis}_headers")()
        if current_headers:
            if len(headers) != len(current_headers):
                raise ValueError(
                    f"Length mismatch: Table has {len(current_headers)} {axis}s, "
                    f"new headers have {len(headers)} elements"
                )
        elif len(headers):
            getattr(self._widget, f"_mgui_set_{axis}_count")(len(headers))

    def _iter_slice(self, slc: slice, axis: int) -> Iterator[int]:
        yield from range(*slc.indices(self.shape[axis]))

    def _get_cell(self, row: int, col: int) -> Any:
        return self._widget._mgui_get_cell(row, col)

    def _set_cell(self, row: int, col: int, value: Any) -> None:
        return self._widget._mgui_set_cell(row, col, value)

    def _get_column(self, col: TblKey, rows: slice = SliceNone) -> list:
        try:
            col_idx = self.column_headers.index(col)
        except ValueError as err:
            raise KeyError(f"{col!r} is not a valid column header") from err
        return [self._get_cell(r, col_idx) for r in self._iter_slice(rows, 0)]

    def _set_column(
        self, col: TblKey, value: Collection, rows: slice = SliceNone
    ) -> None:
        if not isinstance(value, Collection):
            raise TypeError(
                f"value to set column data must be collection. got {type(value)}"
            )
        nrows, ncols = self.shape
        try:
            col_idx = self.column_headers.index(col)
        except ValueError:
            col_idx = ncols
        if col_idx >= ncols:
            # order is important
            new_headers = (*self.column_headers, col)
            self._widget._mgui_set_column_count(ncols + 1)
            # not using column_headers.setter to avoid _check_new_headers call
            self._widget._mgui_set_column_headers(new_headers)

        # TODO: discuss whether it should be an exception if number of rows don't match
        if len(value) > nrows:
            self._widget._mgui_set_row_count(len(value))

        for v, row in zip_longest(value, self._iter_slice(rows, 0)):
            self._set_cell(row, col_idx, v)

    def _del_column(self, col: TblKey) -> None:
        try:
            col_idx = self.column_headers.index(col)
        except ValueError as e:
            raise KeyError(f"{col!r} is not a valid column header") from e
        return self._widget._mgui_remove_column(col_idx)

    def _del_row(self, row: TblKey) -> None:
        try:
            row_idx = self.row_headers.index(row)
        except ValueError as e:
            raise KeyError(f"{row!r} is not a valid row header") from e
        self._del_rowi(row_idx)

    def _del_rowi(self, row: int) -> None:
        self._widget._mgui_remove_row(row)

    def _get_row(self, row: TblKey, cols: slice = SliceNone) -> list:
        """Get row by row header."""
        try:
            row_idx = self.row_headers.index(row)
        except ValueError as e:
            raise KeyError(f"{row!r} is not a valid row header") from e
        return self._get_rowi(row_idx, cols)

    def _get_rowi(self, row: int, cols: slice = SliceNone) -> list:
        """Get row by row index."""
        self._assert_row(row)
        return [self._get_cell(row, c) for c in self._iter_slice(cols, 1)]

    def _set_row(self, row: TblKey, value: Collection, cols: slice = SliceNone) -> None:
        """Set row by row header."""
        try:
            row_idx = self.row_headers.index(row)
        except ValueError as e:
            raise KeyError(f"{row!r} is not a valid row header") from e
        self._set_rowi(row_idx, value, cols)

    def _set_rowi(self, row: int, value: Collection, cols: slice = SliceNone) -> None:
        """Set row by row index."""
        self._assert_row(row)
        for v, col in zip(value, self._iter_slice(cols, 1)):
            self._set_cell(row, col, v)

    def _assert_row(self, row: int) -> int:
        nrows = self._widget._mgui_get_row_count()
        if row >= nrows:
            raise IndexError(
                f"index {row} is out of bounds for table with {nrows} rows."
            )
        return row

    def _assert_col(self, col: int) -> int:
        ncols = self._widget._mgui_get_column_count()
        if col >= ncols:
            raise IndexError(
                f"column {col} is out of bounds for table with {ncols} columns."
            )
        return col

    # #### EXPORT METHODS #####

    def to_dataframe(self) -> pandas.DataFrame:
        """Convert TableData to dataframe."""
        try:
            import pandas

            return pandas.DataFrame(
                self.data.to_list(), self.row_headers, self.column_headers
            )
        except ImportError as e:
            raise ImportError(
                "Must install Pandas to convert to convert Table to DataFrame."
            ) from e

    # fmt: off
    @overload
    def to_dict(self, orient: Literal['dict']) -> dict[TblKey, dict[TblKey, list]]: ...
    @overload
    def to_dict(self, orient: Literal['list']) -> dict[TblKey, list]: ...
    @overload
    def to_dict(self, orient: Literal['split']) -> dict[TblKey, Collection]: ...
    @overload
    def to_dict(self, orient: Literal['records']) -> list[dict[TblKey, Any]]: ...
    @overload
    def to_dict(self, orient: Literal['index']) -> dict[TblKey, dict[TblKey, list]]: ...
    @overload
    def to_dict(self, orient: Literal['series']) -> dict[TblKey, pandas.Series]: ...
    # fmt: on

    def to_dict(self, orient: str = "dict") -> list | dict:
        """Convert the Table to a dictionary.

        The type of the key-value pairs can be customized with the parameters
        (see below).

        Parameters
        ----------
        orient : str {'dict', 'list', 'series', 'split', 'records', 'index'}
            Determines the type of the values of the dictionary.

            - 'dict' (default) : dict like {column -> {index -> value}}
            - 'list' : dict like {column -> [values]}
            - 'split' : dict like
              {'index' -> [index], 'columns' -> [columns], 'data' -> [values]}
            - 'records' : list like
              [{column -> value}, ... , {column -> value}]
            - 'index' : dict like {index -> {column -> value}}
            - 'series' : dict like {column -> Series(values)}

        """
        orient = orient.lower()
        col_head = self.column_headers
        row_head = self.row_headers
        nrows, ncols = self.shape
        if _contains_duplicates(col_head):
            warn(
                "Table column headers are not unique, some columns will be omitted.",
                stacklevel=2,
            )
        if orient == "dict":
            if _contains_duplicates(row_head):
                warn(
                    "Table row headers are not unique, some rows will be omitted.",
                    stacklevel=2,
                )
            return {
                col_head[c]: {row_head[r]: self._get_cell(r, c) for r in range(nrows)}
                for c in range(ncols)
            }
        if orient == "list":
            return dict(self)
        if orient == "split":
            return {"data": self.data.to_list(), "index": row_head, "columns": col_head}
        if orient == "records":
            return [
                {col_head[c]: self._get_cell(r, c) for c in range(ncols)}
                for r in range(nrows)
            ]
        if orient == "index":
            if _contains_duplicates(row_head):
                warn(
                    "Table row headers are not unique, some rows will be omitted.",
                    stacklevel=2,
                )
            return {
                row_head[r]: {col_head[c]: self._get_cell(r, c) for c in range(ncols)}
                for r in range(nrows)
            }
        if orient == "series":
            try:
                from pandas import Series

                return {header: Series(self[header]) for header in col_head}
            except ImportError as e:
                raise ImportError("Must install pandas to use to_dict('series')") from e

        raise ValueError(
            "'orient' argument to 'to_dict' must be one of "
            "('dict', list, 'split', 'records', 'index', 'series)"
        )


class DataView:
    """Object that provides 2D numpy-like indexing for Table data."""

    def __init__(self, obj: Table) -> None:
        self._obj = obj

    def __repr__(self) -> str:
        """Return string representation."""
        return f"<Data for {self._obj!r}>"

    # fmt: off
    @overload
    def __getitem__(self, arg: int) -> list: ...
    @overload
    def __getitem__(self, arg: slice) -> list[list]: ...
    @overload
    def __getitem__(self, arg: tuple[int, int]) -> Any: ...
    @overload
    def __getitem__(self, arg: tuple[int, slice]) -> list: ...
    @overload
    def __getitem__(self, arg: tuple[slice, int]) -> list: ...
    @overload
    def __getitem__(self, arg: tuple[slice, slice]) -> list[list]: ...
    # fmt: on

    def __getitem__(self, idx: IndexKey | tuple[IndexKey, IndexKey]) -> Any:
        """Get index."""
        if isinstance(idx, (int, slice)):
            return self.__getitem__((idx, slice(None)))  # type: ignore
        obj = self._obj
        if isinstance(idx, tuple):
            assert len(idx) == 2, "Table Widget only accepts 2 arguments to __getitem__"
            r_idx, c_idx = idx
            if isinstance(r_idx, int):
                if isinstance(c_idx, int):
                    return obj._get_cell(r_idx, c_idx)
                if isinstance(c_idx, slice):
                    return obj._get_rowi(r_idx, c_idx)
            elif isinstance(r_idx, slice):
                if isinstance(c_idx, int):
                    return obj._get_column(c_idx, r_idx)
                if isinstance(c_idx, slice):
                    return [obj._get_rowi(r, c_idx) for r in obj._iter_slice(r_idx, 0)]
        raise ValueError(f"Not a valid idx for __getitem__ {idx!r}")

    def __setitem__(
        self, idx: IndexKey | tuple[IndexKey, IndexKey], value: Any
    ) -> None:
        """Set index."""
        # TODO: deal with bad shapes
        if isinstance(idx, (int, slice)):
            return self.__setitem__((idx, slice(None)), value)
        if isinstance(idx, tuple):
            assert len(idx) == 2, "Table Widget only accepts 2 arguments to __setitem__"
            r_idx, c_idx = idx
            obj = self._obj
            if isinstance(r_idx, int):
                if isinstance(c_idx, int):
                    return obj._set_cell(r_idx, c_idx, value)
                if isinstance(c_idx, slice):
                    return obj._set_rowi(r_idx, value, c_idx)
            elif isinstance(r_idx, slice):
                # handle extended slices
                if r_idx.step and r_idx.step != 1:
                    # TODO: check value is iterable
                    self._assert_extended_slice(r_idx, len(value))
                if isinstance(c_idx, int):
                    return obj._set_column(c_idx, value, r_idx)
                if isinstance(c_idx, slice):
                    # handle extended slices
                    if c_idx.step and c_idx.step != 1:
                        # TODO: check value is iterable
                        self._assert_extended_slice(c_idx, len(value[0]), axis=1)
                    for v, r in zip(value, obj._iter_slice(r_idx, 0)):
                        obj._set_rowi(r, v, c_idx)
                    return
        raise ValueError(f"Not a valid idx for __setitem__ {idx!r}")

    def __delitem__(self, idx: IndexKey | tuple[IndexKey, IndexKey]) -> None:
        """Get index."""
        if isinstance(idx, (int, slice)):
            return self.__delitem__((idx, slice(None)))
        if isinstance(idx, tuple):
            assert len(idx) == 2, "Table Widget only accepts 2 arguments to __delitem__"
            r_idx, c_idx = idx
            for i in idx:
                if not isinstance(i, int) and i != SliceNone:
                    raise ValueError(f"Can only delete full rows/columns, not {idx!r}")
            obj = self._obj
            if isinstance(r_idx, int):
                if c_idx == SliceNone:
                    return obj._del_rowi(r_idx)
                raise ValueError("Can only delete full rows/columns, not cells")
            elif isinstance(r_idx, slice):
                if isinstance(c_idx, int):
                    return obj._del_column(obj.column_headers[c_idx])
                for r in obj._iter_slice(r_idx, 0):
                    obj._del_rowi(r)
                return
        raise ValueError(f"Not a valid idx for __getitem__ {idx!r}")

    def _assert_extended_slice(self, slc: slice, value_len: int, axis: int = 0) -> None:
        slc_len = _range_len(*slc.indices(self._obj.shape[axis]))
        if slc_len != value_len:
            raise ValueError(
                f"attempt to assign sequence of size {value_len} to "
                f"extended slice of size {slc_len} along axis {axis}"
            )

    def to_numpy(self) -> numpy.ndarray:
        """Return a Numpy representation of the Table.

        Only the values in the Table will be returned, the axes labels will be removed.
        """
        try:
            import numpy

            return numpy.array(self[:])
        except ImportError as e:
            raise ImportError("Cannot convert to numpy without numpy installed") from e

    def to_list(self) -> list[list]:
        """Return table data as a list of lists."""
        return self[:]


def _range_len(start: int, stop: int, step: int) -> int:
    return (stop - start - 1) // step + 1


def _contains_duplicates(X: Iterable[Any]) -> bool:
    seen: set[Any] = set()
    seen_add = seen.add
    return any(x in seen or seen_add(x) for x in X)


def _is_dataframe(obj: object) -> TypeGuard[pandas.DataFrame]:
    pandas = sys.modules.get("pandas")
    return isinstance(obj, pandas.DataFrame) if pandas is not None else False


def _is_numpy_array(obj: object) -> TypeGuard[numpy.ndarray]:
    numpy = sys.modules.get("numpy")
    return isinstance(obj, numpy.ndarray) if numpy is not None else False


def _from_nested_column_dict(data: dict) -> tuple[list[list], list]:
    """Return 2D data and row headers from a dict of nested dicts."""
    _index = {frozenset(i) for i in data.values()}
    if len(_index) > 1:
        try:
            import pandas

            df = pandas.DataFrame(data)
            return df.values, df.index
        except ImportError as err:
            raise ValueError(
                "All row-dicts must have the same keys. "
                "Install pandas for better table-from-dict support."
            ) from err
    # preserve order of keys
    index = []
    for v in data.values():
        index = list(v)
        break

    new_data = [[s[i] for i in index] for s in data.values()]
    return [list(x) for x in zip(*new_data)], index


def _from_dict(data: dict) -> tuple[list[list], list, list]:
    """Return normalized data from dict of array-like or row-dicts.

    logic from pandas.DataFrame.from_dict
    """
    if set(data) == {"data", "index", "columns"}:
        return data["data"], data["index"], data["columns"]
    columns = list(data)
    if isinstance(next(iter(data.values())), dict):
        _data, index = _from_nested_column_dict(data)
    else:
        try:
            _data = [list(x) for x in zip(*data.values())]
        except TypeError as err:
            raise ValueError(
                "All values in the dict must be iterable (e.g. a list)."
            ) from err
        index = []
    return _data, index, columns


def _from_records(data: list[dict[TblKey, Any]]) -> tuple[list[list], list, list]:
    """Return normalized data from a list of column dicts."""
    if not data:
        return [], [], []
    _columns = {frozenset(i) for i in data}
    if len(_columns) > 1:
        try:
            import pandas

            df = pandas.DataFrame(data)
            return df.values, df.index, df.columns
        except ImportError as err:
            raise ValueError(
                "All column-dicts must have the same keys. "
                "Install pandas for better table-from-dict support."
            ) from err
    columns = list(data[0])
    _data = [list(d.values()) for d in data]
    return _data, [], columns


def _validate_table_data(
    data: Collection, index: Sequence | None, column: Sequence | None
) -> None:
    """Make sure data matches shape of index and column."""
    nr = len(data)
    if not nr:
        return None
    try:
        nc = len(data[0])  # type: ignore
    except (TypeError, IndexError):
        nc = 1
    if index is not None and len(index) and len(index) != nr:
        raise ValueError(
            f"Shape of passed values is ({nr}, {nc}), "
            f"headers imply ({len(index)}, {len(column) if column else 1})"
        )
    if column is not None and len(column) and len(column) != nc:
        warn(
            f"Shape of passed values is ({nr}, {nc}), "
            f"headers imply ({len(index) if index else 1}, {len(column)}). "
            "Data will be truncated.",
            stacklevel=2,
        )
    return None



import subprocess
from pathlib import Path

import mkdocs_gen_files
import psygnal

from magicgui import widgets
from magicgui.backends import _ipynb, _qtpy

MAKE_IMAGES = True
BACKENDS = {"qt": _qtpy.widgets, "ipynb": _ipynb.widgets}
WIDGETS_PATH = Path("api/widgets")
IMAGES_PATH = Path("images")
WIDGET_PAGE = """
# {name}

{img_link}

Available in backends: {backends}

## Signals
{signals}

::: magicgui.widgets.{name}
    handler: widget_handler
    options:
        docstring_options:
            warn_unknown_params: false
"""


def _set_dark_mode(dark_mode: bool):
    """Set the system to dark mode or not."""
    cmd = ["osascript", "-l", "JavaScript", "-e"]
    if dark_mode:
        cmd += ["Application('System Events').appearancePreferences.darkMode = true"]
    else:
        cmd += ["Application('System Events').appearancePreferences.darkMode = false"]
    subprocess.run(cmd)


# values just for the sake of making the images for docs
VALS = {
    widgets.Password: "password",
    widgets.FileEdit: "/Users/user/Desktop",
    widgets.FloatRangeSlider: (200, 800),
    widgets.RangeSlider: (200, 800),
    widgets.CheckBox: True,
    widgets.LineEdit: "Some value",
    widgets.TextEdit: "Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
    widgets.Label: "Some label",
    widgets.LiteralEvalLineEdit: "{'a': 1, 'b': 2}",
    widgets.ProgressBar: 420,
    widgets.QuantityEdit: "1 meter",
    widgets.Slider: 42,
    widgets.Table: [{"a": 1, "b": 2}, {"a": 3, "b": 4}],
}


def _add_value(obj: widgets.bases.ValueWidget) -> None:
    if isinstance(obj, widgets.bases.ButtonWidget):
        obj.text = "Click me!"

    if obj.__class__ in VALS:
        obj.value = VALS[obj.__class__]
        return

    if isinstance(obj, widgets.bases.CategoricalWidget):
        obj.choices = ["Choice A", "Choice B", "Choice C"]
        obj.value = "Choice A"

    if isinstance(obj, widgets.Dialog):
        obj.append(widgets.Label(value="Are you sure?"))


def _snap_image(_obj: type, _name: str) -> str:
    from qtpy.QtWidgets import QVBoxLayout, QWidget

    outer = QWidget()
    if _obj is widgets.Container:
        return ""
    if issubclass(_obj, widgets.FunctionGui):
        return ""
    if issubclass(_obj, (widgets.TupleEdit, widgets.ListEdit)):
        wdg = _obj(value=(1, 2, 3))
    else:
        wdg = _obj()
    _add_value(wdg)
    outer.setLayout(QVBoxLayout())
    outer.layout().addWidget(wdg.native)
    pth = IMAGES_PATH / f"{_name}.png"
    with mkdocs_gen_files.open(pth, "wb") as f:
        outer.setMinimumWidth(300)  # turns out this is very important for grab
        outer.grab().save(f.name)
    return f'![{_name} widget](../../{pth}){{ loading=lazy, class="widget-image" }}'


def _get_signals(cls: type[widgets.Widget]) -> list[psygnal.Signal]:
    signals = []
    for name in dir(cls):
        if name.startswith("_"):
            continue
        attr = getattr(cls, name)
        if isinstance(attr, psygnal.Signal):
            signals.append(attr)
    return signals


def main() -> None:
    for name in dir(widgets):
        if name.startswith("_"):
            continue
        cls = getattr(widgets, name)
        if (
            isinstance(cls, type)
            and issubclass(cls, widgets.Widget)
            and cls is not widgets.Widget
            and cls.__name__ == name
        ):
            backends = [
                f"`{key}`" for key, module in BACKENDS.items() if hasattr(module, name)
            ]

            img_link = _snap_image(cls, name) if MAKE_IMAGES else ""
            _signals = _get_signals(cls)

            siglines = []
            for sig in _signals:
                param = sig.signature.parameters
                sigstr = ", ".join([str(p.annotation.__name__) for p in param.values()])
                siglines.append(f"* **`{sig._name}({sigstr})`** - {sig.description}")

            with mkdocs_gen_files.open(WIDGETS_PATH / f"{name}.md", "w") as f:
                md = WIDGET_PAGE.format(
                    name=name,
                    backends=", ".join(backends),
                    img_link=img_link,
                    signals="\n".join(siglines),
                )
                f.write(md)


main()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Tuple, Union

from magicgui.types import Undefined, _Undefined

from ._mixins import _OrientationMixin
from ._ranged_widget import MultiValueRangedWidget, RangedWidget, T

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.widgets import protocols

    from ._widget import WidgetKwargs


class SliderWidget(RangedWidget[T], _OrientationMixin):
    """Widget with a constrained value and orientation. Wraps SliderWidgetProtocol.

    Parameters
    ----------
    value : Any, optional
        The starting value for the widget.
    min : float, optional
        The minimum allowable value, by default 0 (or `value` if `value` is less than 0)
    max : float, optional
        The maximum allowable value, by default 999 (or `value` if `value` is greater
        than 999)
    step : float, optional
        The step size for incrementing the value, by default adaptive step is used
    orientation : str, {'horizontal', 'vertical'}
        The orientation for the slider, by default "horizontal"
    readout : bool, optional
        Whether to show the editable spinbox next to the slider
    tracking : bool, optional
        If tracking is enabled (the default), the slider emits the `changed`
        signal while the slider is being dragged. If tracking is disabled,
        the slider emits the `changed` signal only after the user releases
        the slider.
    bind : Callable[[ValueWidget], Any] | Any, optional
        A value or callback to bind this widget. If provided, whenever
        [`widget.value`][magicgui.widgets.bases.ValueWidget.value] is
        accessed, the value provided here will be returned instead. `bind` may be a
        callable, in which case `bind(self)` will be returned (i.e. your bound callback
        must accept a single parameter, which is this widget instance).
    nullable : bool, optional
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.SliderWidgetProtocol

    def __init__(
        self,
        value: T | _Undefined = Undefined,
        *,
        min: float | _Undefined = Undefined,
        max: float | _Undefined = Undefined,
        step: float | _Undefined | None = Undefined,
        orientation: str = "horizontal",
        readout: bool = True,
        tracking: bool = True,
        bind: T | Callable[[protocols.ValueWidgetProtocol], T] | _Undefined = Undefined,
        nullable: bool = False,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        base_widget_kwargs["backend_kwargs"] = {
            "readout": readout,
            "orientation": orientation,
        }
        super().__init__(
            value=value,  # type: ignore
            min=min,
            max=max,
            step=step,
            bind=bind,  # type: ignore
            nullable=nullable,
            **base_widget_kwargs,
        )
        self.readout = readout
        self.tracking = tracking

    @property
    def tracking(self) -> bool:
        """Return whether slider tracking is enabled.

        If tracking is enabled (the default), the slider emits the changed()
        signal while the slider is being dragged. If tracking is disabled,
        the slider emits the changed() signal only when the user releases
        the slider.
        """
        return self._widget._mgui_get_tracking()

    @tracking.setter
    def tracking(self, value: bool) -> None:
        """Set whether slider tracking is enabled."""
        self._widget._mgui_set_tracking(value)

    @property
    def options(self) -> dict:
        """Return options currently being used in this widget."""
        d = super().options.copy()
        d.update({"orientation": self.orientation, "readout": self._readout})
        return d

    @property
    def readout(self) -> bool:
        """Get visibility state of readout widget."""
        return self._readout

    @readout.setter
    def readout(self, value: bool) -> None:
        """Set visibility state of readout widget."""
        self._readout = value
        self._widget._mgui_set_readout_visibility(value)


class MultiValuedSliderWidget(
    MultiValueRangedWidget[Tuple[Union[int, float], ...]], SliderWidget
):
    """Slider widget that expects a iterable value."""

    _widget: protocols.SliderWidgetProtocol



from dataclasses import dataclass, field
from typing import NamedTuple, Optional

import pytest
from typing_extensions import Annotated, TypedDict

from magicgui.schema._ui_field import UiField, build_widget, get_ui_fields
from magicgui.widgets import Container

EXPECTED = (
    UiField(name="a", type=int, nullable=True),
    UiField(name="b", type=str, description="the b"),
    UiField(name="c", default=0.0, type=float, widget="FloatSlider"),
)


def _assert_uifields(cls, instantiate=True):
    result = tuple(get_ui_fields(cls))
    assert result == EXPECTED
    wdg = build_widget(cls)
    assert isinstance(wdg, Container)
    assert wdg.asdict() == {
        "a": 0,
        "b": "",
        "c": 0.0,
    }
    if instantiate:
        instance = cls(a=1, b="hi")
        assert tuple(get_ui_fields(instance)) == EXPECTED
        wdg2 = build_widget(instance)
        assert isinstance(wdg2, Container)
        assert wdg2.asdict() == {
            "a": 1,
            "b": "hi",
            "c": 0.0,
        }


def test_attrs_descriptor():
    attrs = pytest.importorskip("attrs")

    @attrs.define
    class Foo:
        a: Optional[int]
        b: str = attrs.field(metadata={"description": "the b"})
        c: float = attrs.field(default=0.0, metadata={"widget": "FloatSlider"})

    _assert_uifields(Foo)


def test_dataclass():
    @dataclass
    class Foo:
        a: Optional[int]
        b: str = field(metadata={"description": "the b"})
        c: float = field(default=0.0, metadata={"widget": "FloatSlider"})

    _assert_uifields(Foo)


def test_pydantic():
    pytest.importorskip("pydantic")
    import pydantic.version
    from pydantic import BaseModel, Field

    if pydantic.version.VERSION.startswith("1"):

        class Foo(BaseModel):
            a: Optional[int]
            b: str = Field(description="the b")
            c: float = Field(0.0, widget="FloatSlider")

    else:

        class Foo(BaseModel):
            a: Optional[int]
            b: str = Field(description="the b")
            c: float = Field(0.0, json_schema_extra={"widget": "FloatSlider"})

    _assert_uifields(Foo)


def test_pydantic_dataclass():
    pydantic = pytest.importorskip("pydantic")

    @pydantic.dataclasses.dataclass
    class Foo:
        a: Optional[int]
        b: str = pydantic.Field(description="the b")
        c: float = pydantic.Field(0.0, json_schema_extra={"widget": "FloatSlider"})

    _assert_uifields(Foo)


def test_named_tuple():
    class Foo(NamedTuple):
        a: Optional[int]
        b: Annotated[str, UiField(description="the b")]
        c: Annotated[float, UiField(widget="FloatSlider")] = 0.0

    _assert_uifields(Foo)


def test_typed_dict():
    class Foo(TypedDict):
        a: Optional[int]
        b: Annotated[str, UiField(description="the b")]
        c: Annotated[float, UiField(default=0.0, widget="FloatSlider")]

    # instantiaed TypedDicts are plain dicts that don't remember their annotations
    _assert_uifields(Foo, instantiate=False)


def test_function():
    def foo(
        a: Optional[int],
        b: Annotated[str, UiField(description="the b")],
        c: Annotated[float, UiField(widget="FloatSlider")] = 0.0,
    ): ...

    # makes to sense to instantiate a function
    _assert_uifields(foo, instantiate=False)


def test_annotated():
    class Foo(NamedTuple):
        x: Annotated[float, UiField(default=1)] = 0.0

    with pytest.warns(
        UserWarning, match="Cannot set default value in both type annotation and field"
    ):
        fields = get_ui_fields(Foo)
        assert fields[0].default == 0

    class Foo2(NamedTuple):
        x: Annotated[float, UiField(name="y")] = 0.0

    with pytest.warns(
        UserWarning, match="Cannot set name in both type annotation and field"
    ):
        fields = get_ui_fields(Foo2)
        assert fields[0].name == "x"


def test_annotated_types_lib():
    pytest.importorskip("annotated_types")

    from annotated_types import Ge, Gt, Interval, Le, Len, Lt, MultipleOf, __version__

    from magicgui.schema._ui_field import _uikwargs_from_annotated_type as uikwargs

    at_ver = tuple(int(v) for v in __version__.split("."))

    def assert_eq(annotated_type, expected):
        result = uikwargs(annotated_type)
        assert result.pop("type") is int
        result.pop("_original_annotation")
        assert result == expected

    assert_eq(Annotated[int, Ge(0)], {"minimum": 0})
    assert_eq(Annotated[int, Gt(0)], {"exclusive_minimum": 0})
    assert_eq(Annotated[int, Le(0)], {"maximum": 0})
    assert_eq(Annotated[int, Lt(0)], {"exclusive_maximum": 0})
    L = Len(2, max_exclusive=5) if at_ver < (0, 4) else Len(2, max_length=4)
    assert_eq(Annotated[int, L], {"min_items": 2, "max_items": 4})
    assert_eq(Annotated[int, MultipleOf(2)], {"multiple_of": 2})
    assert_eq(
        Annotated[int, Interval(gt=0, lt=2)],
        {"exclusive_minimum": 0, "exclusive_maximum": 2},
    )
    assert_eq(Annotated[int, Interval(ge=1, le=3)], {"minimum": 1, "maximum": 3})

    if at_ver >= (0, 4):
        from annotated_types import MaxLen, MinLen

        assert_eq(Annotated[int, MinLen(2)], {"min_items": 2})
        assert_eq(Annotated[int, MaxLen(4)], {"max_items": 4})


def test_annotated_types_lib_dataclass():
    pytest.importorskip("annotated_types")

    from annotated_types import Ge, Gt, Interval, Le, Len, Lt, MultipleOf

    @dataclass
    class Foo:
        a: Annotated[int, Ge(1)]
        b: Annotated[int, Gt(1)]
        c: Annotated[int, Le(10)]
        d: Annotated[int, Lt(10)]
        e: Annotated[int, MultipleOf(2)]
        f: Annotated[int, Interval(ge=1, le=5)]
        g: Annotated[list, Len(2, 5)]

    wdg = build_widget(Foo)
    assert wdg.a.min == 1
    assert wdg.b.min == 2
    assert wdg.c.max == 10
    assert wdg.d.max == 9
    assert wdg.e.step == 2
    assert wdg.f.min == 1
    assert wdg.f.max == 5
    # assert wdg.g.min_items == 2  # TODO
    # assert wdg.g.max_items == 5  # TODO


def test_resolved_type():
    f: UiField[int] = UiField(type=Annotated["int", UiField(minimum=0)])
    assert f.resolved_type is int

    f = UiField(type="int")
    assert f.resolved_type is int



"""# matplotlib figure example

Basic example of adding a generic QWidget to a container.

Main lesson: add your QWidget to container.native.layout()
as shown on line 30
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvas

from magicgui import magicgui

x = np.linspace(0, 5, 256)
y = np.linspace(0, 5, 256)[:, np.newaxis]
data = np.sin(x) ** 10 + np.cos(10 + y * x) * np.cos(x)

mpl_fig = plt.figure()
ax = mpl_fig.add_subplot(111)
(line,) = ax.plot(data[123])  # linescan through the middle of the image


@magicgui(position={"widget_type": "Slider", "max": 255}, auto_call=True)
def f(position: int):
    """Function demonstrating magicgui combined with matplotlib."""
    line.set_ydata(data[position])
    line.figure.canvas.draw()


# rather than using the Container.append (`f.append`) ...
# grab the native layout and add the QWidget to it
f.native.layout().addWidget(FigureCanvas(mpl_fig))

f.show(run=True)



from __future__ import annotations

import builtins
from abc import ABC, abstractmethod
from math import ceil, log10
from typing import TYPE_CHECKING, Callable, Iterable, Tuple, TypeVar, Union, cast

from magicgui.types import Undefined, _Undefined

from ._value_widget import ValueWidget

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.widgets import protocols

    from ._widget import WidgetKwargs

T = TypeVar("T", int, float, Tuple[Union[int, float], ...])
DEFAULT_MIN = 0.0
DEFAULT_MAX = 1000.0


class RangedWidget(ValueWidget[T]):
    """Widget with a constrained value. Wraps RangedWidgetProtocol.

    Parameters
    ----------
    value : Any, optional
        The starting value for the widget.
    min : float, optional
        The minimum allowable value, by default 0 (or `value` if `value` is less than 0)
    max : float, optional
        The maximum allowable value, by default 999 (or `value` if `value` is greater
        than 999)
    step : float, optional
        The step size for incrementing the value, by default adaptive step is used
    bind : Callable[[ValueWidget], Any] | Any, optional
        A value or callback to bind this widget. If provided, whenever
        [`widget.value`][magicgui.widgets.bases.ValueWidget.value] is
        accessed, the value provided here will be returned instead. `bind` may be a
        callable, in which case `bind(self)` will be returned (i.e. your bound callback
        must accept a single parameter, which is this widget instance).
    nullable : bool, optional
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.RangedWidgetProtocol

    def __init__(
        self,
        value: T | _Undefined = Undefined,
        *,
        min: float | _Undefined = Undefined,
        max: float | _Undefined = Undefined,
        step: float | _Undefined | None = Undefined,
        bind: T | Callable[[ValueWidget], T] | _Undefined = Undefined,
        nullable: bool = False,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        # value should be set *after* min max is set
        super().__init__(
            bind=bind,  # type: ignore
            nullable=nullable,
            **base_widget_kwargs,
        )

        if step is Undefined or step is None:
            self.step = None
            self._widget._mgui_set_step(1)
        else:
            self.step = cast(float, step)

        self.min, self.max = self._init_range(value, min, max)
        if value not in (Undefined, None):
            self.value = value

    def _init_range(
        self,
        value: T | _Undefined,
        min: float | _Undefined,
        max: float | _Undefined,
    ) -> tuple[float, float]:
        """Return min and max based on given value and arguments.

        If min or max are unset, constrain so the given value is within the range.
        """
        tmp_val: tuple[float | int, ...]
        if value is None or isinstance(value, _Undefined):
            tmp_val = (1,)
        elif isinstance(value, (tuple, list)):
            tmp_val = tuple(1 if v is None else float(v) for v in value)
        else:
            tmp_val = (float(value),)

        new_min: float = (
            cast("float", min)
            if min is not Undefined
            else builtins.min(DEFAULT_MIN, *tmp_val)
        )

        if max is Undefined:
            t = 10.0 ** ceil(log10(builtins.max(0, *tmp_val) + 1))
            new_max = builtins.max(DEFAULT_MAX, t) - 1
        else:
            new_max = cast("float", max)

        return new_min, new_max

    @property
    def options(self) -> dict:
        """Return options currently being used in this widget."""
        d = super().options.copy()
        d.update({"min": self.min, "max": self.max, "step": self.step})
        return d

    @ValueWidget.value.setter  # type: ignore
    def value(self, value: T) -> None:
        """Set widget value, will raise Value error if not within min/max."""
        val: tuple[float, ...] = value if isinstance(value, tuple) else (value,)
        if any(float(v) < self.min or float(v) > self.max for v in val):
            raise ValueError(
                f"value {value} is outside of the allowed range: "
                f"({self.min}, {self.max})"
            )
        ValueWidget.value.fset(self, value)  # type: ignore

    @property
    def min(self) -> float:
        """Minimum allowable value for the widget."""
        return self._widget._mgui_get_min()

    @min.setter
    def min(self, value: float) -> None:
        self._widget._mgui_set_min(value)

    @property
    def max(self) -> float:
        """Maximum allowable value for the widget."""
        return self._widget._mgui_get_max()

    @max.setter
    def max(self, value: float) -> None:
        self._widget._mgui_set_max(value)

    @property
    def step(self) -> float | None:
        """Step size for widget values (None if adaptive step is turned on)."""
        if self._widget._mgui_get_adaptive_step():
            return None
        return self._widget._mgui_get_step()

    @step.setter
    def step(self, value: float | None) -> None:
        if value is None:
            self._widget._mgui_set_adaptive_step(True)
        else:
            self._widget._mgui_set_adaptive_step(False)
            self._widget._mgui_set_step(value)

    @property
    def adaptive_step(self) -> bool:
        """Whether the step size is adaptive.

        Adaptive decimal step means that the step size will continuously be adjusted to
        one power of ten below the current value. So when the value is 1100, the step is
        set to 100, so stepping up once increases it to 1200. For 1200 stepping up takes
        it to 1300. For negative values, stepping down from -1100 goes to -1200.
        """
        return self.step is None

    @adaptive_step.setter
    def adaptive_step(self, value: bool) -> None:
        self.step = None if value else self._widget._mgui_get_step()

    @property
    def range(self) -> tuple[float, float]:
        """Range of allowable values for the widget."""
        return self.min, self.max

    @range.setter
    def range(self, value: tuple[float, float]) -> None:
        self.min, self.max = value


class TransformedRangedWidget(RangedWidget[float], ABC):
    """Widget with a constrained value. Wraps RangedWidgetProtocol.

    This can be used to map one domain of numbers onto another, useful for creating
    things like LogSliders.  Subclasses must reimplement ``_value_from_position`` and
    ``_position_from_value``.

    Parameters
    ----------
    value : Any, optional
        The starting value for the widget.
    min : float, optional
        The minimum allowable value, by default 0
    max : float, optional
        The maximum allowable value, by default 100
    min_pos : float, optional
        The minimum value for the *internal* (widget) position, by default 0.
    max_pos : float, optional
        The maximum value for the *internal* (widget) position, by default 0.
    step : float, optional
        The step size for incrementing the value, by default 1
    bind : Callable[[ValueWidget], Any] | Any, optional
        A value or callback to bind this widget. If provided, whenever
        [`widget.value`][magicgui.widgets.bases.ValueWidget.value] is
        accessed, the value provided here will be returned instead. `bind` may be a
        callable, in which case `bind(self)` will be returned (i.e. your bound callback
        must accept a single parameter, which is this widget instance).
    nullable : bool, optional
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.RangedWidgetProtocol

    def __init__(
        self,
        value: T | _Undefined = Undefined,
        *,
        min: float = 0,
        max: float = 100,
        min_pos: int = 0,
        max_pos: int = 100,
        step: int = 1,
        bind: T | Callable[[ValueWidget], T] | _Undefined = Undefined,
        nullable: bool = False,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        self._min = min
        self._max = max
        self._min_pos = min_pos
        self._max_pos = max_pos
        ValueWidget.__init__(  # type: ignore
            self, value=value, bind=bind, nullable=nullable, **base_widget_kwargs
        )

        self._widget._mgui_set_min(self._min_pos)
        self._widget._mgui_set_max(self._max_pos)
        self._widget._mgui_set_step(step)

    # Just a linear scaling example.
    # Replace _value_from_position, and _position_from_value in subclasses
    # to implement more complex value->position lookups
    @property
    def _scale(self) -> float:
        """Slope of a linear map.  Just used as an example."""
        return (self.max - self.min) / (self._max_pos - self._min_pos)

    @abstractmethod
    def _value_from_position(self, position: float) -> float:
        """Return 'real' value given internal widget position."""
        return self.min + self._scale * (position - self._min_pos)

    @abstractmethod
    def _position_from_value(self, value: float) -> float:
        """Return internal widget position given 'real' value."""
        return (value - self.min) / self._scale + self._min_pos

    #########

    @property
    def value(self) -> float:
        """Return current value of the widget."""
        return self._value_from_position(self._widget._mgui_get_value())

    @value.setter
    def value(self, value: float) -> None:
        return self._widget._mgui_set_value(self._position_from_value(value))

    @property
    def min(self) -> float:
        """Minimum allowable value for the widget."""
        return self._min

    @min.setter
    def min(self, value: float) -> None:
        prev = self.value
        self._min = value
        self.value = prev

    @property
    def max(self) -> float:
        """Maximum allowable value for the widget."""
        return self._max

    @max.setter
    def max(self, value: float) -> None:
        prev = self.value
        self._max = value
        self.value = prev


class MultiValueRangedWidget(RangedWidget[T]):
    """Widget with a constrained *iterable* value, like a tuple."""

    @ValueWidget.value.setter  # type: ignore
    def value(self, value: tuple[float, ...]) -> None:
        """Set widget value, will raise Value error if not within min/max."""
        if not isinstance(value, Iterable):
            raise ValueError(
                f"value {value!r} is not iterable, and must be for a "
                "MultiValueRangedWidget"
            )

        value = tuple(value)
        for v in value:
            if not (self.min <= float(v) <= self.max):
                raise ValueError(
                    f"value {v} is outside of the allowed range: "
                    f"({self.min}, {self.max})"
                )
        ValueWidget.value.fset(self, value)  # type: ignore



"""Functions that map python types to widgets."""

from ._type_map import get_widget_class, register_type, type2callback, type_registered

__all__ = [
    "get_widget_class",
    "register_type",
    "type_registered",
    "type2callback",
]



from __future__ import annotations

from typing import Any, Callable, Hashable, Iterable, Mapping

from magicgui.types import FileDialogMode

from ._concrete import Dialog
from .bases import create_widget


def show_file_dialog(
    mode: str | FileDialogMode = FileDialogMode.EXISTING_FILE,
    caption: str | None = None,
    start_path: str | None = None,
    filter: str | None = None,
    parent: Any = None,
) -> str | None:
    """Immediately show an 'open file' dialog and block.

    Parameters
    ----------
    mode : str or magicgui.types.FileDialogMode, optional
        The mode of the dialog.  Allowed strings are:
        - r: returns one existing file.
        - rm: return one or more existing files.
        - w: return one file name that does not have to exist.
        - d: returns one existing directory.
    caption : str, optional
        Caption for the dialog, by default None
    start_path : str, optional
        Starting directory, by default None
    filter : str, optional
        The filter is used to specify the kind of files that should be shown.
        It should be a glob-style string, like ``'*.png'`` (this may be
        backend-specific), by default None
    parent : widget, optional
        The backend parent widget, by default None

    Returns
    -------
    str or None
        The the path selected, or None if canceled.
    """
    from magicgui import use_app

    app = use_app()
    assert app.native
    func: Callable[..., str | None] = app.get_obj("show_file_dialog")
    return func(mode, caption, start_path, filter, parent)


def request_values(
    values: Mapping | Iterable[tuple[Hashable, Any]] = (),
    *,
    title: str = "",
    parent: Any | None = None,
    **kwargs: type | dict,
) -> dict[str, Any] | None:
    """Show a dialog with a set of values and request the user to enter them.

    Dialog is modal and immediately blocks execution until user closes it.
    If the dialog is accepted, the values are returned as a dictionary, otherwise
    returns `None`.

    See also the docstring of :func:`magicgui.widgets.create_widget` for more
    information.

    Parameters
    ----------
    values : Union[Mapping, Iterable[Tuple[Hashable, Any]]], optional
        A mapping of name to arguments to create_widget.  Values can be a dict, in which
        case they are kwargs to `create_widget`, or a single value, in which case it is
        interpreted as the `annotation` in `create_widget`, by default ()
    title : str
        An optional label to put at the top., by default ""
    parent : Widget, optional
        An optional parent widget, by default None.
        The dialog will inherit style from the parent object.
        CAREFUL: if a parent is set, and subsequently deleted, this widget will likely
        be deleted as well (depending on the backend), and will no longer be usable.
    kwargs : Union[Type, Dict]
        Additional keyword arguments are used as name -> annotation arguments to
        `create_widget`.

    Returns
    -------
    Optional[Dict[str, Any]]
        Dictionary of values if accepted, or `None` if canceled.

    Examples
    --------
    >>> from magicgui.widgets import request_values

    >>> request_values(age=int, name=str, title="Hi! Who are you?")

    >>> request_values(
    ...     age=dict(value=40),
    ...     name=dict(annotation=str, label="Enter your name:"),
    ...     title="Hi! Who are you?"
    ... )

    >>> request_values(
    ...     values={'age': int, 'name': str},
    ...     title="Hi! Who are you?"
    ... )
    """
    widgets = []
    if title:
        from . import Label

        widgets.append(Label(value=title))

    for key, val in dict(values, **kwargs).items():
        kwargs = val if isinstance(val, dict) else {"annotation": val}
        kwargs.setdefault("name", key)
        widgets.append(create_widget(**kwargs))  # type: ignore

    d = Dialog(widgets=widgets, parent=parent)
    return d.asdict() if d.exec() else None



from __future__ import annotations

import inspect
import warnings
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Iterator

from psygnal import Signal, SignalInstance

from magicgui._type_resolution import resolve_single_type
from magicgui.application import Application, use_app
from magicgui.widgets import protocols

if TYPE_CHECKING:
    from weakref import ReferenceType

    import numpy as np
    from typing_extensions import TypedDict

    from magicgui.widgets._concrete import _LabeledWidget
    from magicgui.widgets.protocols import WidgetProtocol

    class WidgetKwargs(TypedDict, total=False):
        # technically, this should be Required[type[WidgetProtocol]]]
        # but the widget_type argument is generally provided dynamically
        # by the @backend_widget decorator. So it appears to be missing if we require it
        widget_type: type[WidgetProtocol]
        name: str
        annotation: Any | None
        label: str | None
        tooltip: str | None
        visible: bool | None
        enabled: bool
        gui_only: bool
        parent: Any
        backend_kwargs: dict | None

        description: str  # alias for label


class Widget:
    """Basic Widget, wrapping a class that implements WidgetProtocol.

    Parameters
    ----------
    widget_type : type[WidgetProtocol]
        A class implementing a widget protocol.  Will be instantiated during __init__.
    name : str, optional
        The name of the parameter represented by this widget. by default ""
    annotation : Any, optional
        The type annotation for the parameter represented by the widget, by default
        ``None``
    label : str
        A string to use for an associated Label widget (if this widget is being
        shown in a [`Container`][magicgui.widgets.Container] widget, and labels are on).
        By default, ``name`` will be used. Note: ``name`` refers the name of the
        parameter, as might be used in a signature, whereas label is just the label
        for that widget in the GUI.
    tooltip : str, optional
        A tooltip to display when hovering over the widget.
    visible : bool, optional
        Whether the widget is visible, by default ``True``.
    enabled : bool, optional
        Whether the widget is enabled, by default ``True``.
    gui_only : bool, optional
        If ``True``, widget is excluded from any function signature representation.
        by default ``False``.  (This will likely be deprecated.)
    parent : Any, optional
        Optional parent widget of this widget.  CAREFUL: if a parent is set, and
        subsequently deleted, this widget will likely be deleted as well (depending on
        the backend), and will no longer be usable.
    backend_kwargs : dict, optional
        keyword argument to pass to the backend widget constructor.
    """

    _widget: WidgetProtocol
    # if this widget becomes owned by a labeled widget
    _labeled_widget_ref: ReferenceType[_LabeledWidget] | None = None

    native_parent_changed = Signal(
        object,
        description="Emitted with the backend widget when the widget parent changes.",
    )
    label_changed = Signal(str, description="Emitted when the widget label changes.")

    def __init__(
        self,
        widget_type: type[WidgetProtocol],
        name: str = "",
        annotation: Any | None = None,
        label: str | None = None,
        tooltip: str | None = None,
        visible: bool | None = None,
        enabled: bool = True,
        gui_only: bool = False,
        parent: Any | None = None,
        backend_kwargs: dict | None = None,
        **extra: Any,  # not really used
    ):
        # for ipywidgets API compatibility
        if backend_kwargs is None:
            backend_kwargs = {}
        label = label or extra.pop("description", None)
        if extra:
            raise TypeError(
                f"{type(self).__name__} got an unexpected "
                f"keyword argument: {', '.join(extra)}"
            )
        for m in self.__class__.__mro__[:-1]:
            _prot = m.__annotations__.get("_widget")
            if _prot:
                break
        else:
            raise TypeError(
                f"Widget type {self.__class__} declared no _widget annotation"
            )
        if not isinstance(_prot, str):
            _prot = _prot.__name__
        prot = getattr(protocols, _prot.replace("protocols.", ""))
        protocols.assert_protocol(widget_type, prot)
        self.__magicgui_app__ = use_app()
        assert self.__magicgui_app__.native
        if isinstance(parent, Widget):
            parent = parent.native
        try:
            self._widget = widget_type(parent=parent, **backend_kwargs)
        except TypeError as e:
            if "unexpected keyword" not in str(e) and "no arguments" not in str(e):
                raise

            warnings.warn(
                "Beginning with magicgui v0.6, the `widget_type` class passed to "
                "`magicgui.Widget` must accept a `parent` Argument. In v0.7 this "
                "will raise an exception. "
                f"Please update '{widget_type.__name__}.__init__()'",
                stacklevel=2,
            )
            self._widget = widget_type(**backend_kwargs)

        self.name: str = name
        self.param_kind = inspect.Parameter.POSITIONAL_OR_KEYWORD
        self._label = label
        self.tooltip = tooltip
        self.enabled = enabled
        self.annotation: Any = annotation
        self.gui_only = gui_only
        self._widget._mgui_bind_parent_change_callback(self.native_parent_changed.emit)

        # put the magicgui widget on the native object...may cause error on some backend
        self.native._magic_widget = self
        self._post_init()
        self._visible: bool = False
        self._explicitly_hidden: bool = False
        if visible is not None:
            self.visible = visible

    @property
    def parent_changed(self) -> SignalInstance:
        """Signal emitted when the parent of the widget changes."""
        warnings.warn(
            "The 'parent_changed' signal has been renamed to 'native_parent_changed'. "
            "Its use will be removed or repurposed in a future version.",
            FutureWarning,
            stacklevel=2,
        )
        return self.native_parent_changed

    @property
    def annotation(self) -> Any:
        """Return type annotation for the parameter represented by the widget.

        ForwardRefs will be resolve when setting the annotation.
        """
        return self._annotation

    @annotation.setter
    def annotation(self, value: Any) -> None:
        self._annotation = resolve_single_type(value)

    @property
    def param_kind(self) -> inspect._ParameterKind:
        """Return :attr:`inspect.Parameter.kind` represented by this widget.

        Used in building signatures from multiple widgets, by default
        :attr:`~inspect.Parameter.POSITIONAL_OR_KEYWORD`
        """
        return self._param_kind

    @param_kind.setter
    def param_kind(self, kind: str | inspect._ParameterKind) -> None:
        if isinstance(kind, str):
            kind = inspect._ParameterKind[kind.upper()]
        if not isinstance(kind, inspect._ParameterKind):
            raise TypeError(
                "'param_kind' must be either a string or a inspect._ParameterKind."
            )
        self._param_kind: inspect._ParameterKind = kind

    def _post_init(self) -> None:
        pass

    @property
    def options(self) -> dict:
        """Return options currently being used in this widget."""
        return {"enabled": self.enabled, "visible": self.visible}

    @property
    def native(self) -> Any:
        """
        Return native backend widget.

        Note this is the widget that contains the layout, and not any
        parent widgets of this (e.g. a parent widget that is used to
        enable scroll bars)
        """
        return self._widget._mgui_get_native_widget()

    @property
    def root_native_widget(self) -> Any:
        """
        Return the root native backend widget.

        This can be different from the ``.native`` widget if the layout
        is a child of some other widget, e.g. a widget used to enable
        scroll bars.
        """
        return self._widget._mgui_get_root_native_widget()

    @property
    def enabled(self) -> bool:
        """Whether widget is enabled (editable)."""
        return self._widget._mgui_get_enabled()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._widget._mgui_set_enabled(value)

    @property
    def parent(self) -> Widget:
        """Return the parent widget."""
        return self._widget._mgui_get_parent()

    @parent.setter
    def parent(self, value: Widget) -> None:
        self._widget._mgui_set_parent(value)

    @property
    def widget_type(self) -> str:
        """Return type of widget."""
        return self.__class__.__name__

    @property
    def label(self) -> str:
        """Return a label to use for this widget when present in Containers."""
        return self.name.replace("_", " ") if self._label is None else self._label

    @label.setter
    def label(self, value: str) -> None:
        self._label = value
        self.label_changed.emit(value)

    @property
    def width(self) -> int:
        """Return the current width of the widget."""
        return self._widget._mgui_get_width()

    @width.setter
    def width(self, value: int) -> None:
        """Set the minimum allowable width of the widget."""
        self._widget._mgui_set_width(value)

    @property
    def min_width(self) -> int:
        """Get the minimum width of the widget."""
        return self._widget._mgui_get_min_width()

    @min_width.setter
    def min_width(self, value: int) -> None:
        """Set the minimum width of the widget."""
        self._widget._mgui_set_min_width(value)

    @property
    def max_width(self) -> int:
        """Get the maximum width of the widget."""
        return self._widget._mgui_get_max_width()

    @max_width.setter
    def max_width(self, value: int) -> None:
        """Set the maximum width of the widget."""
        self._widget._mgui_set_max_width(value)

    @property
    def height(self) -> int:
        """Return the current height of the widget."""
        return self._widget._mgui_get_height()

    @height.setter
    def height(self, value: int) -> None:
        """Set the minimum allowable height of the widget."""
        self._widget._mgui_set_height(value)

    @property
    def min_height(self) -> int:
        """Get the minimum height of the widget."""
        return self._widget._mgui_get_min_height()

    @min_height.setter
    def min_height(self, value: int) -> None:
        """Set the minimum height of the widget."""
        self._widget._mgui_set_min_height(value)

    @property
    def max_height(self) -> int:
        """Get the maximum height of the widget."""
        return self._widget._mgui_get_max_height()

    @max_height.setter
    def max_height(self, value: int) -> None:
        """Set the maximum height of the widget."""
        self._widget._mgui_set_max_height(value)

    @property
    def tooltip(self) -> str | None:
        """Get the tooltip for this widget."""
        return self._widget._mgui_get_tooltip() or None

    @tooltip.setter
    def tooltip(self, value: str | None) -> None:
        """Set the tooltip for this widget."""
        return self._widget._mgui_set_tooltip(value)

    def _labeled_widget(self) -> _LabeledWidget | None:
        """Return _LabeledWidget container, if applicable."""
        return self._labeled_widget_ref() if self._labeled_widget_ref else None

    @property
    def visible(self) -> bool:
        """Return whether widget is visible."""
        return self._widget._mgui_get_visible()

    @visible.setter
    def visible(self, value: bool) -> None:
        """Set widget visibility.

        ``widget.show()`` is an alias for ``widget.visible = True``
        ``widget.hide()`` is an alias for ``widget.visible = False``
        """
        if value is None:
            return

        self._widget._mgui_set_visible(value)
        self._explicitly_hidden = not value

        labeled_widget = self._labeled_widget()
        if labeled_widget is not None:
            labeled_widget.visible = value

    def show(self, run: bool = False) -> Widget:
        """Show widget.

        alias for ``widget.visible = True``

        Parameters
        ----------
        run : bool, optional
            Whether to start the application event loop, by default False
        """
        self.visible = True
        if run:
            self.__magicgui_app__.run()
        return self  # useful for generating repr in sphinx

    @contextmanager
    def shown(self) -> Iterator[Application]:
        """Context manager to show the widget."""
        try:
            self.show()
            yield self.__magicgui_app__.__enter__()
        finally:
            self.__magicgui_app__.__exit__()

    def hide(self) -> None:
        """Hide widget.

        alias for ``widget.visible = False``
        """
        self.visible = False

    def close(self) -> None:
        """Close widget."""
        self._widget._mgui_close_widget()

    def render(self) -> np.ndarray:
        """Return an RGBA (MxNx4) numpy array bitmap of the rendered widget."""
        return self._widget._mgui_render()

    def __repr__(self) -> str:
        """Return representation of widget of instance."""
        return f"{self.widget_type}(annotation={self.annotation!r}, name={self.name!r})"

    def _repr_png_(self) -> bytes | None:
        """Return PNG representation of the widget for QtConsole."""
        from io import BytesIO

        try:
            from imageio import imwrite
        except ImportError:
            print(
                "(For a nicer magicgui widget representation in "
                "Jupyter, please `pip install imageio`)"
            )
            return None

        rendered = self.render()
        if rendered is not None:
            with BytesIO() as file_obj:
                imwrite(file_obj, rendered, format="png")
                file_obj.seek(0)
                return file_obj.read()
        return None

    def _ipython_display_(self, *args: Any, **kwargs: Any) -> Any:
        if hasattr(self.native, "_ipython_display_"):
            return self.native._ipython_display_(*args, **kwargs)
        raise NotImplementedError()

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> dict:
        if hasattr(self.native, "_repr_mimebundle_"):
            return self.native._repr_mimebundle_(*args, **kwargs)  # type: ignore
        raise NotImplementedError()



from __future__ import annotations

import contextlib
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Mapping,
    MutableSequence,
    NoReturn,
    Sequence,
    TypeVar,
    overload,
)

from psygnal import Signal

from magicgui._util import debounce
from magicgui.application import use_app
from magicgui.signature import MagicParameter, MagicSignature, magic_signature
from magicgui.widgets.bases._mixins import _OrientationMixin

from ._button_widget import ButtonWidget
from ._value_widget import ValueWidget
from ._widget import Widget

WidgetVar = TypeVar("WidgetVar", bound=Widget)

if TYPE_CHECKING:
    import inspect
    from pathlib import Path

    from typing_extensions import Unpack

    from magicgui.widgets import Container, protocols

    from ._widget import WidgetKwargs

    class ContainerKwargs(WidgetKwargs, Generic[WidgetVar], total=False):
        widgets: Sequence[WidgetVar]
        layout: str
        scrollable: bool
        labels: bool


class ContainerWidget(Widget, _OrientationMixin, MutableSequence[WidgetVar]):
    """Widget that can contain other widgets.

    Wraps a widget that implements
    [`ContainerProtocol`][magicgui.widgets.protocols.ContainerProtocol].

    A `ContainerWidget` behaves like a python list of [Widget][magicgui.widgets.Widget]
    objects. Subwidgets can be accessed using integer or slice-based indexing
    (`container[0]`), as well as by widget name (`container.<widget_name>`). Widgets can
    be added with `append` or `insert`, and removed with `del` or `pop`, etc...

    There is a tight connection between a `ContainerWidget` and an
    [inspect.Signature][inspect.Signature] object,
    just as there is a tight connection between individual [Widget` objects an
    an :class:`inspect.Parameter][inspect.Parameter] object.
    The signature representation of a `ContainerWidget`
    (with the current settings as default values) is accessible with
    the :meth:`~ContainerWidget.__signature__` method (or by using
    :func:`inspect.signature` from the standard library)

    For a `ContainerWidget` subclass that is tightly coupled to a specific function
    signature (as in the "classic" magicgui decorator), see
    [magicgui.widgets.FunctionGui][magicgui.widgets.FunctionGui].

    Parameters
    ----------
    widgets : Sequence[Widget], optional
        A sequence of widgets with which to initialize the container, by default
        `None`.
    layout : str, optional
        The layout for the container.  must be one of `{'horizontal',
        'vertical'}`. by default "vertical"
    scrollable : bool, optional
        Whether to enable scroll bars or not. If enabled, scroll bars will
        only appear along the layout direction, not in both directions.
    labels : bool, optional
        Whether each widget should be shown with a corresponding Label widget to the
        left, by default `True`.  Note: the text for each widget defaults to
        `widget.name`, but can be overridden by setting `widget.label`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    changed = Signal(
        object,
        description="Emitted with `self` when any sub-widget in the container changes.",
    )
    _widget: protocols.ContainerProtocol
    _initialized = False
    # this is janky ... it's here to allow connections during __init__ by
    # avoiding a recursion error in __getattr__
    _list: list[WidgetVar] = []  # noqa: RUF012

    def __init__(
        self,
        widgets: Sequence[WidgetVar] = (),
        *,
        layout: str = "vertical",
        scrollable: bool = False,
        labels: bool = True,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        self._list: list[WidgetVar] = []
        self._labels = labels
        self._layout = layout
        self._scrollable = scrollable
        base_widget_kwargs.setdefault("backend_kwargs", {}).update(  # type: ignore
            {"layout": layout, "scrollable": scrollable}
        )
        super().__init__(**base_widget_kwargs)
        self.extend(widgets)
        self.native_parent_changed.connect(self.reset_choices)
        self._initialized = True
        self._unify_label_widths()

    def __getattr__(self, name: str) -> WidgetVar:
        """Return attribute ``name``.  Will return a widget if present."""
        for widget in self._list:
            if name == widget.name:
                return widget
        return object.__getattribute__(self, name)  # type: ignore

    def __setattr__(self, name: str, value: Any) -> None:
        """Set attribute ``name``.  Prevents changing widget if present, (use del)."""
        if self._initialized:
            for widget in self._list:
                if name == widget.name:
                    raise AttributeError(
                        "Cannot set attribute with same name as a widget\n"
                        "If you are trying to change the value of a widget, use: "
                        f"`{self.__class__.__name__}.{name}.value = {value}`",
                    )
        object.__setattr__(self, name, value)

    @overload
    def __getitem__(self, key: int | str) -> WidgetVar: ...

    @overload
    def __getitem__(self, key: slice) -> MutableSequence[WidgetVar]: ...

    def __getitem__(
        self, key: int | str | slice
    ) -> WidgetVar | MutableSequence[WidgetVar]:
        """Get item by integer, str, or slice."""
        if isinstance(key, str):
            return self.__getattr__(key)
        if isinstance(key, slice):
            return [getattr(item, "_inner_widget", item) for item in self._list[key]]
        elif isinstance(key, int):
            item = self._list[key]
            return getattr(item, "_inner_widget", item)
        raise TypeError(f"list indices must be integers or slices, not {type(key)}")

    def index(self, value: Any, start: int = 0, stop: int = 9223372036854775807) -> int:
        """Return index of a specific widget instance (or widget name)."""
        if isinstance(value, str):
            value = getattr(self, value)
        return super().index(value, start, stop)

    def remove(self, value: Widget | str) -> None:
        """Remove a widget instance (may also be string name of widget)."""
        super().remove(value)  # type: ignore

    def __delattr__(self, name: str) -> None:
        """Delete a widget by name."""
        self.remove(name)

    def __delitem__(self, key: int | slice) -> None:
        """Delete a widget by integer or slice index."""
        if isinstance(key, slice):
            for item in self._list[key]:
                ref = getattr(item, "_labeled_widget_ref", None)
                if ref:
                    item = ref()
                self._widget._mgui_remove_widget(item)
        elif isinstance(key, int):
            item = self._list[key]
            ref = getattr(item, "_labeled_widget_ref", None)
            if ref:
                item = item._labeled_widget_ref()  # type: ignore
            self._widget._mgui_remove_widget(item)
        else:
            raise TypeError(f"list indices must be integers or slices, not {type(key)}")
        del self._list[key]

    def __len__(self) -> int:
        """Return the count of widgets."""
        return len(self._list)

    def __setitem__(self, key: Any, value: Any) -> NoReturn:
        """Prevent assignment by index."""
        raise RuntimeError("magicgui.Container does not support item setting.")

    def __dir__(self) -> list[str]:
        """Add subwidget names to the dir() call for this widget."""
        d = list(super().__dir__())
        d.extend([w.name for w in self._list if not w.gui_only])
        return d

    def insert(self, key: int, widget: WidgetVar) -> None:
        """Insert widget at ``key``."""
        if isinstance(widget, (ValueWidget, ContainerWidget)):
            widget.changed.connect(lambda: self.changed.emit(self))
        _widget = widget

        if self.labels:
            from magicgui.widgets._concrete import _LabeledWidget

            # no labels for button widgets (push buttons, checkboxes, have their own)
            if not isinstance(widget, (_LabeledWidget, ButtonWidget)):
                _widget = _LabeledWidget(widget)  # type: ignore
                widget.label_changed.connect(self._unify_label_widths)

        if key < 0:
            key += len(self)
        self._list.insert(key, widget)
        # NOTE: if someone has manually mucked around with self.native.layout()
        # it's possible that indices will be off.
        self._widget._mgui_insert_widget(key, _widget)
        self._unify_label_widths()

    def _unify_label_widths(self) -> None:
        if not self._initialized:
            return

        need_labels = [w for w in self._list if not isinstance(w, ButtonWidget)]
        if self.layout == "vertical" and self.labels and need_labels:
            measure = use_app().get_obj("get_text_width")
            widest_label = max(measure(w.label) for w in need_labels)
            for w in self:
                labeled_widget = w._labeled_widget()
                if labeled_widget:
                    labeled_widget.label_width = widest_label

    @property
    def margins(self) -> tuple[int, int, int, int]:
        """Return margin between the content and edges of the container."""
        return self._widget._mgui_get_margins()

    @margins.setter
    def margins(self, margins: tuple[int, int, int, int]) -> None:
        # left, top, right, bottom
        self._widget._mgui_set_margins(margins)

    @property
    def layout(self) -> str:
        """Return the layout of the widget."""
        return self._layout

    @layout.setter
    def layout(self, value: str) -> NoReturn:
        raise NotImplementedError(
            "It is not yet possible to change layout after instantiation"
        )

    def reset_choices(self, *_: Any) -> None:
        """Reset choices for all Categorical subWidgets to the default state.

        If widget._default_choices is a callable, this may NOT be the exact same set of
        choices as when the widget was instantiated, if the callable relies on external
        state.
        """
        for widget in self._list:
            if hasattr(widget, "reset_choices"):
                widget.reset_choices()

    @property
    def __signature__(self) -> MagicSignature:
        """Return a MagicSignature object representing the current state of the gui."""
        params = [
            MagicParameter.from_widget(w)
            for w in self._list
            if w.name and not w.gui_only
        ]
        # if we have multiple non-default parameters and some but not all of them are
        # "bound" to fallback values, we may have  non-default arguments
        # following default arguments
        seen_default = False
        for p in params:
            if p.default is not p.empty:
                seen_default = True
            elif seen_default:
                params.sort(key=lambda x: x.default is not MagicParameter.empty)
                break
        return MagicSignature(params)

    @classmethod
    def from_signature(
        cls, sig: inspect.Signature, **kwargs: Unpack[ContainerKwargs]
    ) -> Container:
        """Create a Container widget from an inspect.Signature object."""
        return MagicSignature.from_signature(sig).to_container(**kwargs)

    @classmethod
    def from_callable(
        cls,
        obj: Callable,
        gui_options: dict | None = None,
        **kwargs: Unpack[ContainerKwargs],
    ) -> Container:
        """Create a Container widget from a callable object.

        In most cases, it will be preferable to create a ``FunctionGui`` instead.
        """
        return magic_signature(obj, gui_options=gui_options).to_container(**kwargs)

    def __repr__(self) -> str:
        """Return a repr."""
        return f"<Container {self.__signature__}>"

    @property
    def labels(self) -> bool:
        """Whether widgets are presented with labels."""
        return self._labels

    @labels.setter
    def labels(self, value: bool) -> None:
        if value == self._labels:
            return
        self._labels = value

        for index, _ in enumerate(self):
            widget = self.pop(index)
            self.insert(index, widget)

    NO_VALUE = "NO_VALUE"

    def asdict(self) -> dict[str, Any]:
        """Return state of widget as dict."""
        return {
            w.name: getattr(w, "value", None)
            for w in self._list
            if w.name and not w.gui_only
        }

    def update(
        self,
        mapping: Mapping | Iterable[tuple[str, Any]] | None = None,
        **kwargs: Any,
    ) -> None:
        """Update the parameters in the widget from a mapping, iterable, or kwargs."""
        with self.changed.blocked():
            if mapping:
                items = mapping.items() if isinstance(mapping, Mapping) else mapping
                for key, value in items:
                    getattr(self, key).value = value
            for key, value in kwargs.items():
                getattr(self, key).value = value
        self.changed.emit()

    @debounce
    def _dump(self, path: str | Path) -> None:
        """Dump the state of the widget to `path`."""
        import pickle
        from pathlib import Path

        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        _dict = {}
        for widget in self._list:
            try:
                # not all values will be pickleable and restorable...
                # for now, don't even try
                _v = pickle.dumps(getattr(widget, "value", self.NO_VALUE))
                _dict[widget.name] = _v
            except Exception:
                continue

        path.write_bytes(pickle.dumps(_dict))

    def _load(self, path: str | Path, quiet: bool = False) -> None:
        """Restore the state of the widget from previously saved file at `path`."""
        import pickle
        from pathlib import Path

        path = Path(path)
        if not path.exists():
            if quiet:
                return
            raise FileNotFoundError(f"Widget state file does not exist: {path}")

        try:
            data: dict = pickle.loads(path.read_bytes())
        except Exception:
            if quiet:
                path.unlink(missing_ok=True)
                return
            raise

        for key, val in data.items():
            with contextlib.suppress(ValueError, AttributeError):
                val = pickle.loads(val)
                if val != self.NO_VALUE:
                    getattr(self, key).value = val


class MainWindowWidget(ContainerWidget):
    """Top level Application widget that can contain other widgets."""

    _widget: protocols.MainWindowProtocol

    def create_menu_item(
        self,
        menu_name: str,
        item_name: str,
        callback: Callable | None = None,
        shortcut: str | None = None,
    ) -> None:
        """Create a menu item ``item_name`` under menu ``menu_name``.

        ``menu_name`` will be created if it does not already exist.
        """
        self._widget._mgui_create_menu_item(menu_name, item_name, callback, shortcut)


class DialogWidget(ContainerWidget):
    """Modal Container."""

    _widget: protocols.DialogProtocol

    def exec(self) -> bool:
        """Show the dialog, and block.

        Return True if the dialog was accepted, False if rejected.
        """
        return bool(self._widget._mgui_exec())



"""`guiclass` decorator.

1. Turns a class into a dataclass
2. Uses `psygnal.evented` to make it an [evented
   dataclass](https://psygnal.readthedocs.io/en/latest/dataclasses/)
2. Adds a `gui` property to the class that will return a `magicgui` widget, bound to
the values of the dataclass instance.
"""

from __future__ import annotations

import contextlib
import warnings
from dataclasses import Field, dataclass, field, is_dataclass
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Mapping, TypeVar, overload

from psygnal import SignalGroup, SignalInstance, evented
from psygnal import __version__ as psygnal_version

from magicgui.schema._ui_field import build_widget
from magicgui.widgets import PushButton
from magicgui.widgets.bases import ContainerWidget, ValueWidget

if TYPE_CHECKING:
    from typing import Protocol

    from typing_extensions import TypeGuard

    # fmt: off
    class GuiClassProtocol(Protocol):
        """Protocol for a guiclass."""

        @property
        def gui(self) -> ContainerWidget: ...
        @property
        def events(self) -> SignalGroup: ...
    # fmt: on

__all__ = ["guiclass", "button", "is_guiclass", "unbind_gui_from_instance"]
_BUTTON_ATTR = "_magicgui_button"
_GUICLASS_FLAG = "__magicgui_guiclass__"

_T = TypeVar("_T")
T = TypeVar("T", bound="type[Any]")
F = TypeVar("F", bound=Callable)


if psygnal_version.split(".")[:2] < ["0", "8"]:
    _IGNORE_REF_ERR = {}
else:
    _IGNORE_REF_ERR = {"on_ref_error": "ignore"}


# https://github.com/microsoft/pyright/blob/main/specs/dataclass_transforms.md
def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_specifiers: tuple[type | Callable[..., Any], ...] = (()),
) -> Callable[[_T], _T]:
    return lambda a: a


@__dataclass_transform__(field_specifiers=(Field, field))
@overload
def guiclass(cls: T) -> T: ...


@__dataclass_transform__(field_specifiers=(Field, field))
@overload
def guiclass(
    *,
    gui_name: str = "gui",
    events_namespace: str = "events",
    follow_changes: bool = True,
    **dataclass_kwargs: Any,
) -> Callable[[T], T]: ...


def guiclass(
    cls: T | None = None,
    *,
    gui_name: str = "gui",
    events_namespace: str = "events",
    follow_changes: bool = True,
    **dataclass_kwargs: Any,
) -> T | Callable[[T], T]:
    """Turn class into a dataclass with a property (`gui_name`) that returns a gui.

    This decorator is similar to `dataclasses.dataclass`, but it will also add an
    `events` attribute to the class that is an instance of `psygnal.SignalGroup` (with a
    signal for each field in the dataclass; see
    https://psygnal.readthedocs.io/en/latest/dataclasses/ for details), and a `gui`
    property that returns a `magicgui` widget, bound to the values of the dataclass
    instance.

    !!! note
        This decorator is compatible with dataclasses using `slots=True`, however,
        there is a potential for a memory leak that the user should be aware of.
        If you create a `guiclass` instance, and then store a reference to its `gui`,
        and then delete the instance, the `gui` will still be bound to the instance,
        preventing it from being garbage collected.  To avoid this, you can call
        `unbind_gui_from_instance(gui, instance)` before deleting the instance.

    Parameters
    ----------
    cls : type
        The class to turn into a dataclass.
    gui_name : str, optional
        The name of the property that will return a `magicgui` widget, by default
        `"gui"`
    events_namespace : str, optional
        The name of the attribute that will be added to the class, by default "events".
        This attribute will be an instance of `psygnal.SignalGroup` that will be used
        to connect events to the class.
    follow_changes : bool, optional
        If `True` (default), changes to the dataclass instance will be reflected in the
        gui, and changes to the gui will be reflected in the dataclass instance.
    dataclass_kwargs : dict, optional
        Additional keyword arguments to pass to `dataclasses.dataclass`.

    Returns
    -------
    type
        The dataclass.

    Examples
    --------
    >>> @guiclass
    ... class MyData:
    ...     x: int = 0
    ...     y: str = 'hi'
    ...
    ...     @button
    ...     def reset(self):
    ...         self.x = 0
    ...         self.y = 'hi'
    ...
    >>> data = MyData()
    >>> data.gui.show()
    """

    def _deco(cls: T) -> T:
        if dataclass_kwargs.get("frozen", False):
            raise ValueError(
                "The 'guiclass' decorator does not support dataclasses with "
                "`frozen=True`. If you need this feature, please open an issue at "
                "https://github.com/pyapp-kit/magicgui/issues."
            )

        setattr(cls, gui_name, GuiBuilder(gui_name, follow_changes=follow_changes))

        if not is_dataclass(cls):
            cls = dataclass(cls, **dataclass_kwargs)  # type: ignore
        cls = evented(cls, events_namespace=events_namespace)

        setattr(cls, _GUICLASS_FLAG, True)
        return cls

    return _deco(cls) if cls is not None else _deco


def is_guiclass(obj: object) -> TypeGuard[GuiClassProtocol]:
    """Return `True` if obj is a guiclass or an instance of a guiclass."""
    return is_dataclass(obj) and hasattr(obj, _GUICLASS_FLAG)


@overload
def button(func: F) -> F: ...


@overload
def button(**kwargs: Any) -> Callable[[F], F]: ...


def button(func: F | None = None, **button_kwargs: Any) -> F | Callable[[F], F]:
    """Add a method as a button to a `guiclass`, which calls the decorated method.

    Parameters
    ----------
    func : callable
        The method to decorate.  If None, returns a decorator that can be applied to a
        method.
    button_kwargs : dict, optional
        Additional keyword arguments to pass to `magicgui.widgets.PushButton`.
    """

    def _deco(func: F) -> F:
        button_kwargs.setdefault("name", func.__name__)
        setattr(func, _BUTTON_ATTR, button_kwargs)
        return func

    return _deco(func) if func else _deco


class GuiBuilder:
    """Descriptor that builds a widget for a dataclass or instance."""

    def __init__(self, name: str = "", follow_changes: bool = True):
        self._name = name
        self._follow_changes = follow_changes

    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name

    def __get__(
        self, instance: object | None, owner: type
    ) -> ContainerWidget[ValueWidget]:
        wdg = build_widget(owner if instance is None else instance)

        # look for @button-decorated methods
        # TODO: move inside build_widget?
        for k, v in vars(owner).items():
            if hasattr(v, _BUTTON_ATTR):
                kwargs = getattr(v, _BUTTON_ATTR)
                button = PushButton(**kwargs)
                if instance is not None:
                    # call the bound method if we're in an instance
                    button.clicked.connect(getattr(instance, k))
                else:
                    # TODO: what to do here?
                    # probably doesn't make sense to call the method on the class
                    ...
                # TODO: allow inserting in a specific location
                wdg.append(button)

        if instance is not None and self._name:
            if self._follow_changes:
                bind_gui_to_instance(wdg, instance)
            # cache it on the instance
            # call del instance.<_name> to remove it
            # this may fail in cases of __slots__ other attribute restrictions
            # the cost is that the widget will be rebuilt on each access
            with contextlib.suppress(AttributeError):
                setattr(instance, self._name, wdg)
        return wdg


def bind_gui_to_instance(
    gui: ContainerWidget, instance: Any, two_way: bool = True
) -> None:
    """Set change events in `gui` to update the corresponding attributes in `model`.

    Parameters
    ----------
    gui : ContainerWidget
        The widget to bind to the instance.
    instance : Any
        The instance to bind to the widget.  In most cases, this will be an instance
        of a `guiclass`.
    two_way : bool, optional
        If True, changes to the instance will be reflected in the gui, by default True
    """
    events = getattr(instance, "events", None) if two_way else None
    signals: Mapping[str, SignalInstance] = {}
    if isinstance(events, SignalGroup):
        # psygnal >=0.10
        if hasattr(events, "__iter__") and hasattr(events, "__getitem__"):
            signals = {k: events[k] for k in events}
        else:  # psygnal <0.10
            signals = events.signals

    # in cases of classes with `__slots__`, widget.changed.connect_setattr(instance...)
    # will show a RuntimeWarning because `instance` will not be weak referenceable.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)

        for widget in gui:
            # would be cleaner to check isinstance(widget, ValueWidget)... but a few
            # widgets (like FileEdit) are not ValueWidgets, but still have a `changed`
            # signal and a value method. So for now we just check for the protocol.
            changed = getattr(widget, "changed", None)

            # we skip PushButton here, otherwise we will set the value of the
            # button (a boolean) to some attribute on the instance, which is probably
            # not what we want.
            if isinstance(changed, SignalInstance) and not isinstance(
                widget, PushButton
            ):
                name: str = getattr(widget, "name", "")
                # connect changes in the widget to the instance
                if hasattr(instance, name):
                    try:
                        changed.connect_setattr(
                            instance,
                            name,
                            maxargs=1,
                            **_IGNORE_REF_ERR,  # type: ignore
                        )
                    except TypeError:
                        warnings.warn(
                            f"Could not bind {name} to {instance}. "
                            "This may be because the instance has __slots__ or "
                            "other attribute restrictions. Please update psygnal.",
                            stacklevel=2,
                        )

                # connect changes from the instance to the widget
                if name in signals:
                    signals[name].connect_setattr(widget, "value")


def unbind_gui_from_instance(gui: ContainerWidget, instance: Any) -> None:
    """Unbind a gui from an instance.

    This will disconnect all events that were connected by `bind_gui_to_instance`.
    In cases where `guiclass` is used on a dataclass with `slots=True`, when a reference
    to the gui is stored elsewhere, this must be called by the user before the instance
    is deleted.

    Parameters
    ----------
    gui : ContainerWidget
        The widget bound to `instance`.
    instance : Any
        An instance of a `guiclass`.
    """
    for widget in gui:
        if isinstance(widget, ValueWidget):
            widget.changed.disconnect_setattr(instance, widget.name, missing_ok=True)


# Class-based form ... which provides subclassing and inheritance (unlike @guiclass)


@__dataclass_transform__(field_specifiers=(Field, field))
class GuiClassMeta(type):
    def __new__(cls: type, name: str, bases: tuple, attrs: dict) -> type:
        attrs[_GUICLASS_FLAG] = True
        obj: type = type.__new__(cls, name, bases, attrs)
        return evented(dataclass(obj))


# evented will warn "No mutable fields found in class <class '__main__.GuiClass'>"
# no events will be emitted... but it will work fine for subclasses.
with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=UserWarning)

    class GuiClass(metaclass=GuiClassMeta):
        gui = GuiBuilder()
        if TYPE_CHECKING:
            events: ClassVar[SignalGroup]

            # the mypy dataclass magic doesn't work without the literal decorator
            # it WILL work with pyright due to the __dataclass_transform__ above
            # here we just avoid a false error in mypy
            def __init__(self, *args: Any, **kwargs: Any) -> None: ...



"""Backend modules implementing applications and widgets."""

from __future__ import annotations

BACKENDS: dict[str, tuple[str, str]] = {
    "Qt": ("_qtpy", "qtpy"),
    "ipynb": ("_ipynb", "ipynb"),
}

for key in list(BACKENDS):
    if not key.islower():
        BACKENDS[key.lower()] = BACKENDS.pop(key)



"""# Password login

A password login field widget.
"""

from magicgui import magicgui


# note that "password" is a special keyword argument
# it will create a password field in the gui by default
# (unless you override "widget_type")
# whereas "password2" will be a normal text field
# (unless you override "widget_type")
@magicgui(password2={"widget_type": "Password"})
def login(username: str, password: str, password2: str):
    """User login credentials."""
    ...


login.show(run=True)



"""MagicSignature objects are an extension to inspect.Signature objects.

The basic idea is that there is a tight connection between an individual Parameter
in a function signature and an individual magicgui.Widget, and a connection between
a full function Signature (a collection of parameters) and a magicgui.Container
(a collection of widgets).  It should be easy to go from function signature to Container
(with ``MagicSignature.to_container()``) and from Container to signature (using
``ContainerWidget.to_signature()``).

calling ``inspect.signature`` on a function decorated with `magicgui` still works,
(it returns a ``MagicSignature``, which is a subclass of ``inspect.Signature``)

"""

from __future__ import annotations

import inspect
import typing
import warnings
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Callable, Sequence, cast

from typing_extensions import Annotated, get_args, get_origin

from magicgui.types import Undefined

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.application import AppRef
    from magicgui.widgets import Container, Widget
    from magicgui.widgets.bases._container_widget import ContainerKwargs

TZ_EMPTY = "__no__default__"


def make_annotated(annotation: Any = Any, options: dict | None = None) -> Any:
    """Merge a annotation and an options dict into an Annotated type.

    Parameters
    ----------
    annotation : Any, optional
        Type Annotation, by default Any
    options : dict, optional
        widget options, by default None

    Returns
    -------
    Annotated
        Annotated type with form ``Annotated[annotation, options]``

    Raises
    ------
    TypeError
        If options is provided and is not a dict
        If ``annotation`` is an Annotated option but the metadata is not a dict
        If ``annotation`` is not a valid type.
    """
    if options and not isinstance(options, dict):
        raise TypeError("'options' must be a dict")
    _options = (options or {}).copy()

    if get_origin(annotation) is Annotated:
        # nested Annotated type results in multiple dicts.
        annotation, *anno_options = get_args(annotation)
        for opt in anno_options:
            try:
                _opt = dict(opt)
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Every item in Annotated must be castable to a dict, got: {opt!r}"
                ) from e
            _options.update(_opt)

    annotation = _make_hashable(annotation)
    _options = _make_hashable(_options)
    return Annotated[annotation, _options]


# this is a stupid hack to deal with something that changed in typing-extensions
# v4.12.0 (and probably python 3.13), where all items in Annotated must now be hashable
# The PR that necessitated this was https://github.com/python/typing_extensions/pull/392


class hashabledict(dict):
    """Hashable immutable dict."""

    def __hash__(self) -> int:  # type: ignore # noqa: D105
        # we don't actually want to hash the contents, just the object itself.
        return id(self)

    def __setitem__(self, key: Any, value: Any) -> None:  # noqa: D105
        warnings.warn(
            "Mutating magicgui Annotation metadata after creation is not supported."
            "This will raise an error in a future version.",
            stacklevel=2,
        )
        super().__setitem__(key, value)

    def __delitem__(self, key: Any) -> None:  # noqa: D105
        raise TypeError("hashabledict is immutable")


def _hashable(obj: Any) -> bool:
    try:
        hash(obj)
        return True
    except TypeError:
        return False


def _make_hashable(obj: Any) -> Any:
    if _hashable(obj):
        return obj
    if isinstance(obj, dict):
        return hashabledict({k: _make_hashable(v) for k, v in obj.items()})
    if isinstance(obj, (list, tuple)):
        return tuple(_make_hashable(v) for v in obj)
    if isinstance(obj, set):
        return frozenset(_make_hashable(v) for v in obj)
    if (args := get_args(obj)) and (not _hashable(args)):
        try:
            obj = get_origin(obj)[_make_hashable(args)]
        except TypeError:
            # python 3.8 hack
            if obj.__module__ == "typing" and hasattr(obj, "_name"):
                generic = getattr(typing, obj._name)
                return generic[_make_hashable(args)]
            raise
    return obj


class _void:
    """private sentinel."""


class MagicParameter(inspect.Parameter):
    """A Parameter subclass that is closely linked to a magicgui.Widget object.

    Parameters
    ----------
    name : str, optional
        The name of the parameter represented by this widget. by default ""
    kind : inspect._ParameterKind, optional
        The :attr:`inspect.Parameter.kind` represented by this widget.  Used in building
        signatures from multiple widgets, by default "POSITIONAL_OR_KEYWORD"
    default : Any, optional
        The default value for the parameter, by default None
    annotation : Any, optional
        The type annotation for the parameter represented by the widget, by default
        None
    options : dict, optional
        Dict of options to pass to the Widget constructor, by default dict()

    """

    def __init__(
        self,
        name: str,
        kind: inspect._ParameterKind = inspect.Parameter.POSITIONAL_OR_KEYWORD,
        *,
        default: Any = inspect.Parameter.empty,
        annotation: Any = inspect.Parameter.empty,
        gui_options: dict | None = None,
        raise_on_unknown: bool = False,
    ):
        _annotation = make_annotated(annotation, gui_options)
        super().__init__(name, kind, default=default, annotation=_annotation)
        self.raise_on_unknown = raise_on_unknown

    @property
    def options(self) -> dict:
        """Return just this options part of the annotation."""
        return cast(dict, get_args(self.annotation)[1])

    def __repr__(self) -> str:
        """Return __repr__, replacing NoneType if present."""
        rep = f"{super().__repr__()[:-1]} {self.options}>"
        rep = rep.replace(": NoneType = ", "=")
        return rep

    def __str__(self) -> str:
        """Return string representation of the Parameter in a signature."""
        hint, _ = get_args(self.annotation)
        return str(
            inspect.Parameter(
                self.name, self.kind, default=self.default, annotation=hint
            )
        )

    def to_widget(self, app: AppRef | None = None) -> Widget:
        """Create and return a widget for this object."""
        from magicgui.widgets import create_widget

        value = Undefined if self.default in (self.empty, TZ_EMPTY) else self.default
        widget = create_widget(
            name=self.name,
            value=value,
            annotation=self.annotation,
            app=app,
            raise_on_unknown=self.raise_on_unknown,
        )
        widget.param_kind = self.kind
        return widget

    @classmethod
    def from_widget(cls, widget: Widget) -> MagicParameter:
        """Create a MagicParameter object representing a widget."""
        return cls(
            name=str(widget.name),
            kind=widget.param_kind,
            default=getattr(widget, "value", inspect.Parameter.empty),
            annotation=widget.annotation,
            gui_options=widget.options,
        )

    @classmethod
    def from_parameter(
        cls,
        param: inspect.Parameter,
        gui_options: dict | None = None,
        raise_on_unknown: bool = False,
    ) -> MagicParameter:
        """Create MagicParameter from an inspect.Parameter."""
        if isinstance(param, MagicParameter):
            return param
        return cls(
            param.name,
            param.kind,
            default=param.default,
            annotation=param.annotation,
            gui_options=gui_options,
            raise_on_unknown=raise_on_unknown,
        )


class MagicSignature(inspect.Signature):
    """A Signature subclass that is closely linked to a magicgui.Container object.

    Parameters
    ----------
    parameters : Optional[Sequence[inspect.Parameter]], optional
        A list of parameters to include in the Signature, by default None
    return_annotation : Type or str, optional
        An optional return annotation to use when representing this container of
        widgets as an inspect.Signature, by default inspect.Signature.empty
    gui_options : Optional[Dict[str, dict]], optional
        A mapping of parameter name to options for each individual parameter,
        by default None
    """

    parameters: MappingProxyType[str, MagicParameter]

    def __init__(
        self,
        parameters: Sequence[inspect.Parameter] | None = None,
        *,
        return_annotation: Any = inspect.Signature.empty,
        gui_options: dict[str, dict] | None = None,
        raise_on_unknown: bool = False,
    ) -> None:
        params = [
            MagicParameter.from_parameter(
                p, (gui_options or {}).get(p.name), raise_on_unknown
            )
            for p in parameters or []
        ]
        super().__init__(params, return_annotation=return_annotation)

    @classmethod
    def from_signature(
        cls,
        sig: inspect.Signature,
        gui_options: dict | None = None,
        raise_on_unknown: bool = False,
    ) -> MagicSignature:
        """Convert regular inspect.Signature to MagicSignature."""
        if type(sig) is cls:
            return sig
        elif not isinstance(sig, inspect.Signature):
            raise TypeError("'sig' must be an instance of 'inspect.Signature'")
        return cls(
            list(sig.parameters.values()),
            return_annotation=sig.return_annotation,
            gui_options=gui_options,
            raise_on_unknown=raise_on_unknown,
        )

    def widgets(self, app: AppRef | None = None) -> MappingProxyType:
        """Return mapping from parameters to widgets for all params in Signature."""
        return MappingProxyType(
            {n: p.to_widget(app) for n, p in self.parameters.items()}
        )

    def to_container(
        self, app: AppRef | None = None, **kwargs: Unpack[ContainerKwargs]
    ) -> Container:
        """Return a ``magicgui.widgets.Container`` for this MagicSignature."""
        from magicgui.widgets import Container

        kwargs["widgets"] = list(self.widgets(app).values())
        return Container(**kwargs)

    def replace(  # type: ignore[override]
        self,
        *,
        parameters: Sequence[inspect.Parameter] | type[_void] = _void,
        return_annotation: Any = _void,
    ) -> MagicSignature:
        """Create a customized copy of the Signature.

        Pass ``parameters`` and/or ``return_annotation`` arguments
        to override them in the new copy.
        """
        if parameters is _void:
            parameters = list(self.parameters.values())

        if return_annotation is _void:
            return_annotation = self.return_annotation

        return type(self)(
            cast(Sequence[inspect.Parameter], parameters),
            return_annotation=return_annotation,
        )


def magic_signature(
    obj: Callable,
    *,
    gui_options: dict[str, dict] | None = None,
    follow_wrapped: bool = True,
    raise_on_unknown: bool = False,
) -> MagicSignature:
    """Create a MagicSignature from a callable object.

    This is magicgui's equivalent of ``inspect.signature`` and is a core
    part of the FunctionGui creation when using the `@magicgui` decorator.

    Parameters
    ----------
    obj : Callable
        The function being inspected.
    gui_options : Optional[Dict[str, dict]], optional
        A dict of name: widget_options dict for each parameter in the function.
        Will be passed to `MagicSignature.from_signature` by default None
    follow_wrapped : bool, optional
        passed to inspect.signature, by default True
    raise_on_unknown : bool, optional
        If True, raise an error if a parameter annotation is not recognized.

    Returns
    -------
    MagicSignature
        Signature object representing the callable.

    Raises
    ------
    ValueError
        If ``gui_options`` are provided with keyword arguments that do not match
        parameters in the decorated function.
    TypeError
        If any of the values in ``gui_options`` are not dicts.
    """
    sig = inspect.signature(obj, follow_wrapped=follow_wrapped)
    if gui_options:
        invalid = set(gui_options) - set(sig.parameters)
        if invalid:
            raise ValueError(
                f"Received parameter option key(s) {invalid} that do not match "
                f"parameters in the provided function: {sig}"
            )
        bad = {v for v in gui_options.values() if not isinstance(v, dict)}
        if bad:
            s = "s" if len(bad) > 1 else ""
            raise TypeError(f"Value for parameter{s} {bad} must be a dict")

    return MagicSignature.from_signature(
        sig, gui_options=gui_options, raise_on_unknown=raise_on_unknown
    )
